import numpy as np
import multiprocessing as mp
from mrl.replays.core.replay_buffer import RingBuffer
from mrl.utils.misc import AttrDict
from multiprocessing import RawArray, RawValue
from collections import OrderedDict

BUFF = None

def worker_init(buffer : AttrDict):
  global BUFF
  BUFF = buffer

def future_samples(idxs):
  """Assumes there is an 'ag' field, and samples n transitions, pairing each with a future ag
  from its trajectory."""
  global BUFF

  # get original transitions
  transition = []
  for buf in BUFF.items:
    item = BUFF[buf].get_batch(idxs)
    transition.append(item)

  # add random future goals
  tlefts = BUFF.buffer_tleft.get_batch(idxs)
  idxs = idxs + np.round(np.random.uniform(size=len(idxs)) * tlefts).astype(np.int64)
  ags = BUFF.buffer_ag.get_batch(idxs)
  transition.append(ags)
  return transition


def actual_samples(idxs):
  """Assumes there is an 'dg' field, and samples n transitions, pairing each with a random dg"""
  global BUFF

  # get original transitions
  transition = []
  for buf in BUFF.items:
    item = BUFF[buf].get_batch(idxs)
    transition.append(item)

  # add random actual goals
  idxs = np.random.choice(len(BUFF.buffer_tidx), len(idxs))
  dgs = BUFF.buffer_dg.get_batch(idxs)
  transition.append(dgs)

  return transition


def achieved_samples(idxs):
  """Assumes there is an 'ag' field, and samples n transitions, pairing each with a random ag"""
  global BUFF

  # get original transitions
  transition = []
  for buf in BUFF.items:
    item = BUFF[buf].get_batch(idxs)
    transition.append(item)

  # add random achieved goals
  idxs = np.random.choice(len(BUFF.buffer_tidx), len(idxs))
  ags = BUFF.buffer_ag.get_batch(idxs)
  transition.append(ags)

  return transition


def behavioral_samples(idxs):
  """Assumes there is an 'bg' field, and samples n transitions, pairing each with a random bg"""
  global BUFF

  # get original transitions
  transition = []
  for buf in BUFF.items:
    item = BUFF[buf].get_batch(idxs)
    transition.append(item)

  # add random achieved goals
  idxs = np.random.choice(len(BUFF.buffer_tidx), len(idxs))
  bgs = BUFF.buffer_bg.get_batch(idxs)
  transition.append(bgs)

  return transition


def n_step_samples(args):
  """Samples s_t and s_{t+n}, along with the discounted reward in between.
  Assumes buffers include: state, action, reward, next_state, done
  Because some sampled states will not have enough future states, this will
  sometimes return less than num_samples samples.
  """
  state_idxs, n_steps, gamma = args

  global BUFF

  tlefts = BUFF.buffer_tleft.get_batch(state_idxs)

  # prune idxs for which there are not enough future transitions
  good_state_idxs = state_idxs[tlefts >= n_steps - 1]
  potentially_bad_idxs = state_idxs[tlefts < n_steps - 1] # 0 tleft corresp to 1 step
  potentially_bad_delt = tlefts[tlefts < n_steps - 1]
  also_good_state_idxs = potentially_bad_idxs[np.isclose(BUFF.buffer_done.get_batch(potentially_bad_idxs + potentially_bad_delt), 1).reshape(-1)]

  state_idxs = np.concatenate((good_state_idxs, also_good_state_idxs), axis=0)
  t_idxs = BUFF.buffer_tidx.get_batch(state_idxs)

  # start building the transition, of state, action, n_step reward, n_step next_state, done
  transition = [BUFF.buffer_state.get_batch(state_idxs), BUFF.buffer_action.get_batch(state_idxs)]
  rewards = np.zeros_like(state_idxs, dtype=np.float32).reshape(-1, 1)

  for i in range(n_steps):
    query = state_idxs + i
    r_delta = (gamma ** i) * BUFF.buffer_reward.get_batch(query)
    diff_traj = t_idxs != BUFF.buffer_tidx.get_batch(query)
    r_delta[diff_traj] *= 0.
    rewards += r_delta

  transition.append(rewards)
  transition.append(BUFF.buffer_next_state.get_batch(query)) # n_step state

  dones = BUFF.buffer_done.get_batch(query)
  dones += diff_traj.astype(np.float32).reshape(-1, 1)
  transition.append(dones)

  return transition



# def stacked_samples(args):
#   """Samples the state and next state as a sequence of states."""
#   num_samples, num_previous_states = args

#   global BUFF
#   base_idxs = np.random.choice(len(BUFF.buffer_tidx), n)
#   idxs = [base_idxs] + [base_idxs - (i + 1) for i in range(num_previous_states)]

#   # get original transitions
#   transition = []
#   for buf in BUFF.items:
#     if buf == 'buffer_state':
#       item =
#     item = BUFF[buf].get_batch(idxs)
#     transition.append(item)

#   # add random future goals
#   tlefts = BUFF.buffer_tleft.get_batch(idxs)
#   idxs = idxs + np.round(np.random.uniform(size=n) * tlefts).astype(np.int64)
#   ags = BUFF.buffer_ag.get_batch(idxs)
#   transition.append(ags)
#   return transition

class SharedMemoryTrajectoryBuffer():
  """
  An alternative replay buffer object that stores contiguous trajectories in a shared memory buffer.
  This should allow for expensive queries (like online hindsight relabeling) to be executed in parallel.

  NOTE: even hindsight relabeling is only faster with Multiprocessing when using minibatches of 5000+.
  E.g.:
    - for sample size 10K:, sample_actual = 1.36ms vs 2.34 (MP vs single proc); future = 1.69ms vs 2.31ms
    - for sample size 500:, sample_actual = 420us vs 220us (MP vs single proc); future = 390us vs 160us

  So probably should just leave it on n_cpu = 1, unless there is something really expensive to do, like
  computing statistics of whole trajectories, or searching forward through the buffer.
  """

  def __init__(self, limit, item_shape, n_cpu=1):
    """
    The replay buffer object. Stores everything in float32.

    :param limit: (int) the max number of transitions to store
    :param item_shape: a list of tuples of (str) item name and (tuple) the shape for item
      Ex: [("observations", env.observation_space.shape),\
          ("actions",env.action_space.shape),\
          ("rewards", (1,)),\
          ("dones", (1,))]
    """
    self.limit = limit

    global BUFF
    BUFF = AttrDict()
    self.BUFF = BUFF # a global object that has shared RawArray-based RingBuffers.

    BUFF.items = []

    # item buffers
    for name, shape in item_shape:
      BUFF.items.append('buffer_' + name)
      BUFF['raw_' + name] = RawArray('f', int(np.prod((limit, ) + shape)))
      BUFF['np_' + name] =\
        np.frombuffer(BUFF['raw_' + name], dtype=np.float32).reshape((limit, ) + shape)
      BUFF['buffer_' + name] = RingBuffer(limit, shape=shape, data=BUFF['np_' + name])

    # special buffers
    BUFF.raw_tidx = RawArray('d', limit)
    BUFF.np_tidx = np.frombuffer(BUFF.raw_tidx, dtype=np.int64)
    BUFF.buffer_tidx = RingBuffer(limit, shape=(), dtype=np.int64, data=BUFF.np_tidx)

    BUFF.raw_tleft = RawArray('d', limit)
    BUFF.np_tleft = np.frombuffer(BUFF.raw_tleft, dtype=np.int64)
    BUFF.buffer_tleft = RingBuffer(limit, shape=(), dtype=np.int64, data=BUFF.np_tleft)

    if 'buffer_bg' in BUFF: # is this a successful trajectory?
      BUFF.raw_success = RawArray('f', limit)
      BUFF.np_success = np.frombuffer(BUFF.raw_success, dtype=np.float32)
      BUFF.buffer_success = RingBuffer(limit, shape=(), dtype=np.float32, data=BUFF.np_success)

    self.trajectories = OrderedDict() # a centralized dict of trajectory_id --> trajectory_idxs
    self.total_trajectory_len = 0
    self.current_trajectory = 0

    self.pool = None
    self.n_cpu = n_cpu
    if n_cpu > 1:
      self.pool = mp.Pool(n_cpu, initializer=worker_init, initargs=(BUFF,))

  def add_trajectory(self, *items):
    """
    Append a trajectory of transitions to the buffer.

    :param items: a list of batched transition values to append to the replay buffer,
        in the item order that we initialized the ReplayBuffer with.
    """
    trajectory_len = len(items[0])

    for buffer_name, batched_values in zip(self.BUFF.items, items):
      self.BUFF[buffer_name].append_batch(batched_values)

    # add to special buffers
    self.BUFF.buffer_tleft.append_batch(np.arange(trajectory_len, dtype=np.int64)[::-1])
    idxs = self.BUFF.buffer_tidx.append_batch(np.ones((trajectory_len,), dtype=np.int64) * self.current_trajectory)

    # Keep track of successes, if has behavioral goals.
    if 'buffer_bg' in BUFF:
      for buffer_name, batched_values in zip(self.BUFF.items, items):
        if buffer_name == 'buffer_reward':
          success = np.any(np.isclose(batched_values, 0.))
          if success:
            self.BUFF.buffer_success.append_batch(np.ones((trajectory_len,), dtype=np.float32))
          else:
            self.BUFF.buffer_success.append_batch(np.zeros((trajectory_len,), dtype=np.float32))

    self.trajectories[self.current_trajectory] = idxs
    self.total_trajectory_len += trajectory_len

    # remove trajectories until all remaining trajectories fit in the buffer.
    while self.total_trajectory_len > self.limit:
      _, idxs = self.trajectories.popitem(last=False)
      self.total_trajectory_len -= len(idxs)
    self.current_trajectory += 1

  def sample_trajectories(self, n, group_by_buffer=False, from_m_most_recent=None):
    """
    Samples n full trajectories (optionally from 'from_m_most_recent' trajectories)
    """

    if from_m_most_recent is not None and (len(self.trajectories) - n >= 0):
      min_idx = max(self.current_trajectory - len(self.trajectories), self.current_trajectory - from_m_most_recent)
      idxs = np.random.randint(min_idx, self.current_trajectory, n)
    else:
      idxs = np.random.randint(self.current_trajectory - len(self.trajectories), self.current_trajectory, n)
    queries = [self.trajectories[i] for i in idxs]
    splits = np.cumsum([len(q) for q in queries[:-1]])
    query = np.concatenate(queries)
    transition = []
    for buf in BUFF.items:
      transition.append(np.split(BUFF[buf].get_batch(query), splits))

    if group_by_buffer:
      return transition

    return list(zip(*transition))

  def sample(self, batch_size, batch_idxs=None):
    """
    sample a random batch from the buffer

    :param batch_size: (int) the number of element to sample for the batch
    :return: (list) the sampled batch
    """
    if self.size == 0:
      return []

    if batch_idxs is None:
      batch_idxs = np.random.randint(self.size, size=batch_size)

    transition = []
    for buf in BUFF.items:
      item = BUFF[buf].get_batch(batch_idxs)
      transition.append(item)

    return transition

  def sample_slices(self, batch_size, slice_size):
    """Tries to sample slices of length slice_size randomly. Slices must be
    from same trajectory, which may not happen even when oversampled, so it's possible
    (but unlikely, at least for small slice_size) to get a small batch size than requested.
    """
    if self.size == 0:
      return [[] for _ in range(slice_size)]

    b_idxs = np.random.randint(self.size, size=int(batch_size * 1.5))
    first_t = BUFF.buffer_tidx.get_batch(b_idxs - slice_size + 1)
    last_t  = BUFF.buffer_tidx.get_batch(b_idxs)

    batch_idxs = b_idxs[first_t == last_t][:batch_size]

    transitions = []
    for i in range(-slice_size + 1, 1):
      transitions.append([])
      for buf in BUFF.items:
        item = BUFF[buf].get_batch(batch_idxs + i)
        transitions[-1].append(item)

    return transitions

  def sample_n_step_transitions(self, batch_size, n_steps, gamma, batch_idxs=None):
    if batch_idxs is None:
      batch_idxs = np.random.randint(self.size, size=batch_size)
      
    if self.pool is not None:
      res = self.pool.map(n_step_samples, zip(np.array_split(batch_idxs, self.n_cpu), [n_steps] * self.n_cpu, [gamma] * self.n_cpu))
      res = [np.concatenate(x, 0) for x in zip(*res)]
      return res

    return n_step_samples((batch_idxs, n_steps, gamma))

  def sample_future(self, batch_size, batch_idxs=None):
    if batch_idxs is None:
      batch_idxs = np.random.randint(self.size, size=batch_size)

    if self.pool is not None:
      res = self.pool.map(future_samples, np.array_split(batch_idxs, self.n_cpu))
      res = [np.concatenate(x, 0) for x in zip(*res)]
      return res

    return future_samples(batch_idxs)

  def sample_from_goal_buffer(self, buffer, batch_size, batch_idxs=None):
    """buffer is one of 'ag', 'dg', 'bg'"""
    if buffer == 'ag':
      sample_fn = achieved_samples
    elif buffer == 'dg':
      sample_fn = actual_samples
    elif buffer == 'bg':
      sample_fn = behavioral_samples
    else:
      raise NotImplementedError

    if batch_idxs is None:
      batch_idxs = np.random.randint(self.size, size=batch_size)

    if self.pool is not None:
      res = self.pool.map(sample_fn, np.array_split(batch_idxs, self.n_cpu))
      res = [np.concatenate(x, 0) for x in zip(*res)]
      return res

    return sample_fn(batch_idxs)

  def __len__(self): return self.size

  def _get_state(self): 
    d = dict(
      trajectories = self.trajectories,
      total_trajectory_len = self.total_trajectory_len,
      current_trajectory = self.current_trajectory
    )
    for bufname in self.BUFF.items + ['buffer_tleft', 'buffer_tidx', 'buffer_success']:
      if bufname in self.BUFF:
        d[bufname] = self.BUFF[bufname]._get_state()
    return d

  def _set_state(self, d):
    for k in ['trajectories','total_trajectory_len','current_trajectory']:
      self.__dict__[k] = d[k]
    for bufname in self.BUFF.items + ['buffer_tleft', 'buffer_tidx', 'buffer_success']:
      if bufname in self.BUFF:
        self.BUFF[bufname]._set_state(*d[bufname])

  def close(self):
    if self.pool is not None:
      self.pool.close()

  @property
  def size(self):
    return len(BUFF.buffer_tidx)

  @property
  def num_trajectories(self):
    return len(self.trajectories)



import numpy as np
from coral_pytorch.losses import corn_loss


def remove_trailing_zeros(arr):
    """
    Remove trailing zeros from array (flatten if needed)
    :param arr: input array
    :return: array without trailing zeros
    """

    # Flatten
    if isinstance(arr, np.ndarray):
        if len(arr.shape) > 1:
            arr = arr.flatten()

    # Iterate
    for idx, n in enumerate(arr[:-1]):

        # Check if the rest is zero too
        if n == 0:
            if not np.count_nonzero(arr[idx + 1:]) > 0:
                return np.copy(arr[:idx])

    return arr


def remove_nones(arr):
    """
    Remove nones from array.
    """
    
    return arr[arr != np.array(None)]


class EnsembleLoss:

    def __init__(self, n_classes):
        self.n_classes = n_classes

    def __call__(self, x, y):
        return corn_loss(x, y, self.n_classes)



"""
  In soft gradient boosting, all base estimators could be simultaneously
  fitted, while achieving the similar boosting improvements as in gradient
  boosting.
"""


import abc
import torch
import logging
import warnings
import torch.nn as nn
import torch.nn.functional as F
from joblib import Parallel, delayed

from ._base import BaseModule, BaseClassifier, BaseRegressor
from ._base import torchensemble_model_doc
from .utils import io
from .utils import set_module
from .utils import operator as op
from .utils.logging import get_tb_logger


__all__ = ["SoftGradientBoostingClassifier", "SoftGradientBoostingRegressor"]


__model_doc = """
    Parameters
    ----------
    estimator : torch.nn.Module
        The class or object of your base estimator.

        - If :obj:`class`, it should inherit from :mod:`torch.nn.Module`.
        - If :obj:`object`, it should be instantiated from a class inherited
          from :mod:`torch.nn.Module`.
    n_estimators : int
        The number of base estimators in the ensemble.
    estimator_args : dict, default=None
        The dictionary of hyper-parameters used to instantiate base
        estimators. This parameter will have no effect if ``estimator`` is a
        base estimator object after instantiation.
    shrinkage_rate : float, default=1
        The shrinkage rate used in gradient boosting.
    cuda : bool, default=True

        - If ``True``, use GPU to train and evaluate the ensemble.
        - If ``False``, use CPU to train and evaluate the ensemble.
    n_jobs : int, default=None
        The number of workers for training the ensemble. This input
        argument is used for parallel ensemble methods such as
        :mod:`voting` and :mod:`bagging`. Setting it to an integer larger
        than ``1`` enables ``n_jobs`` base estimators to be trained
        simultaneously.

    Attributes
    ----------
    estimators_ : torch.nn.ModuleList
        An internal container that stores all fitted base estimators.
"""


__fit_doc = """
    Parameters
    ----------
    train_loader : torch.utils.data.DataLoader
        A data loader that contains the training data.
    epochs : int, default=100
        The number of training epochs per base estimator.
    use_reduction_sum : bool, default=True
        Whether to set ``reduction="sum"`` for the internal mean squared
        error used to fit each base estimator.
    log_interval : int, default=100
        The number of batches to wait before logging the training status.
    test_loader : torch.utils.data.DataLoader, default=None
        A data loader that contains the evaluating data.

        - If ``None``, no validation is conducted after each base
          estimator being trained.
        - If not ``None``, the ensemble will be evaluated on this
          dataloader after each base estimator being trained.
    save_model : bool, default=True
        Specify whether to save the model parameters.

        - If test_loader is ``None``, the ensemble containing
          ``n_estimators`` base estimators will be saved.
        - If test_loader is not ``None``, the ensemble with the best
          validation performance will be saved.
    save_dir : string, default=None
        Specify where to save the model parameters.

        - If ``None``, the model will be saved in the current directory.
        - If not ``None``, the model will be saved in the specified
          directory: ``save_dir``.
"""


def _soft_gradient_boosting_model_doc(header, item="model"):
    """
    Decorator on obtaining documentation for different gradient boosting
    models.
    """

    def get_doc(item):
        """Return the selected item"""
        __doc = {"model": __model_doc, "fit": __fit_doc}
        return __doc[item]

    def adddoc(cls):
        doc = [header + "\n\n"]
        doc.extend(get_doc(item))
        cls.__doc__ = "".join(doc)
        return cls

    return adddoc


def _parallel_compute_pseudo_residual(
    output, target, estimator_idx, shrinkage_rate, n_outputs, is_classification
):
    """
    Compute pseudo residuals in soft gradient boosting for each base estimator
    in a parallel fashion.
    """
    accumulated_output = torch.zeros_like(output[0], device=output[0].device)
    for i in range(estimator_idx):
        accumulated_output += shrinkage_rate * output[i]

    # Classification
    if is_classification:
        residual = op.pseudo_residual_classification(
            target, accumulated_output, n_outputs
        )
    # Regression
    else:
        residual = op.pseudo_residual_regression(target, accumulated_output)

    return residual


class _BaseSoftGradientBoosting(BaseModule):
    def __init__(
        self,
        estimator,
        n_estimators,
        estimator_args=None,
        shrinkage_rate=1.0,
        cuda=True,
        n_jobs=None,
    ):
        super(BaseModule, self).__init__()
        self.base_estimator_ = estimator
        self.n_estimators = n_estimators
        self.estimator_args = estimator_args

        if estimator_args and not isinstance(estimator, type):
            msg = (
                "The input `estimator_args` will have no effect since"
                " `estimator` is already an object after instantiation."
            )
            warnings.warn(msg, RuntimeWarning)

        self.shrinkage_rate = shrinkage_rate
        self.device = torch.device("cuda" if cuda else "cpu")
        self.n_jobs = n_jobs
        self.logger = logging.getLogger()
        self.tb_logger = get_tb_logger()

        self.estimators_ = nn.ModuleList()
        self.use_scheduler_ = False

    def _validate_parameters(self, epochs, log_interval):
        """Validate hyper-parameters on training the ensemble."""

        if not epochs > 0:
            msg = (
                "The number of training epochs = {} should be strictly"
                " positive."
            )
            self.logger.error(msg.format(epochs))
            raise ValueError(msg.format(epochs))

        if not log_interval > 0:
            msg = (
                "The number of batches to wait before printting the"
                " training status should be strictly positive, but got {}"
                " instead."
            )
            self.logger.error(msg.format(log_interval))
            raise ValueError(msg.format(log_interval))

        if not 0 < self.shrinkage_rate <= 1:
            msg = (
                "The shrinkage rate should be in the range (0, 1], but got"
                " {} instead."
            )
            self.logger.error(msg.format(self.shrinkage_rate))
            raise ValueError(msg.format(self.shrinkage_rate))

    @abc.abstractmethod
    def _evaluate_during_fit(self, test_loader, epoch):
        """Evaluate the ensemble after each training epoch."""

    def fit(
        self,
        train_loader,
        epochs=100,
        use_reduction_sum=True,
        log_interval=100,
        test_loader=None,
        save_model=True,
        save_dir=None,
    ):

        # Instantiate base estimators and set attributes
        for _ in range(self.n_estimators):
            self.estimators_.append(self._make_estimator())
        self._validate_parameters(epochs, log_interval)
        self.n_outputs = self._decide_n_outputs(train_loader)

        # Utils
        criterion = (
            nn.MSELoss(reduction="sum") if use_reduction_sum else nn.MSELoss()
        )
        total_iters = 0

        # Set up optimizer and learning rate scheduler
        optimizer = set_module.set_optimizer(
            self, self.optimizer_name, **self.optimizer_args
        )

        if self.use_scheduler_:
            scheduler = set_module.set_scheduler(
                optimizer,
                self.scheduler_name,
                **self.scheduler_args  # noqa: E501
            )

        for epoch in range(epochs):
            self.train()
            for batch_idx, elem in enumerate(train_loader):

                data, target = io.split_data_target(elem, self.device)
                output = [estimator(*data) for estimator in self.estimators_]

                # Compute pseudo residuals in parallel
                rets = Parallel(n_jobs=self.n_jobs)(
                    delayed(_parallel_compute_pseudo_residual)(
                        output,
                        target,
                        i,
                        self.shrinkage_rate,
                        self.n_outputs,
                        self.is_classification,
                    )
                    for i in range(self.n_estimators)
                )

                # Compute sGBM loss
                loss = torch.tensor(0.0, device=self.device)
                for idx, estimator in enumerate(self.estimators_):
                    loss += criterion(output[idx], rets[idx])

                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

                # Print training status
                if batch_idx % log_interval == 0:
                    with torch.no_grad():
                        msg = "Epoch: {:03d} | Batch: {:03d} | RegLoss: {:.5f}"
                        self.logger.info(msg.format(epoch, batch_idx, loss))
                        if self.tb_logger:
                            self.tb_logger.add_scalar(
                                "sGBM/Train_Loss", loss, total_iters
                            )
                total_iters += 1

            # Validation
            if test_loader:
                flag = self._evaluate_during_fit(test_loader, epoch)
                if save_model and flag:
                    io.save(self, save_dir, self.logger)

            # Update the scheduler
            if self.use_scheduler_:
                scheduler.step()

        if save_model and not test_loader:
            io.save(self, save_dir, self.logger)


@_soft_gradient_boosting_model_doc(
    """Implementation on the SoftGradientBoostingClassifier.""", "model"
)
class SoftGradientBoostingClassifier(
    _BaseSoftGradientBoosting, BaseClassifier
):
    def __init__(
        self,
        estimator,
        n_estimators,
        estimator_args=None,
        shrinkage_rate=1.0,
        cuda=True,
        n_jobs=None,
    ):
        super().__init__(
            estimator=estimator,
            n_estimators=n_estimators,
            estimator_args=estimator_args,
            shrinkage_rate=shrinkage_rate,
            cuda=cuda,
            n_jobs=n_jobs,
        )
        self._criterion = nn.CrossEntropyLoss()
        self.is_classification = True
        self.best_acc = 0.0

    @torch.no_grad()
    def _evaluate_during_fit(self, test_loader, epoch):
        self.eval()
        correct = 0
        total = 0
        flag = False
        for _, elem in enumerate(test_loader):
            data, target = io.split_data_target(elem, self.device)
            output = self.forward(*data)
            _, predicted = torch.max(output.data, 1)
            correct += (predicted == target).sum().item()
            total += target.size(0)
        acc = 100 * correct / total

        if acc > self.best_acc:
            self.best_acc = acc
            flag = True

        msg = (
            "Epoch: {:03d} | Validation Acc: {:.3f}"
            " % | Historical Best: {:.3f} %"
        )
        self.logger.info(msg.format(epoch, acc, self.best_acc))
        if self.tb_logger:
            self.tb_logger.add_scalar(
                "soft_gradient_boosting/Validation_Acc", acc, epoch
            )

        return flag

    @torchensemble_model_doc(
        """Set the attributes on optimizer for SoftGradientBoostingClassifier.""",  # noqa: E501
        "set_optimizer",
    )
    def set_optimizer(self, optimizer_name, **kwargs):
        super().set_optimizer(optimizer_name, **kwargs)

    @torchensemble_model_doc(
        """Set the attributes on scheduler for SoftGradientBoostingClassifier.""",  # noqa: E501
        "set_scheduler",
    )
    def set_scheduler(self, scheduler_name, **kwargs):
        super().set_scheduler(scheduler_name, **kwargs)

    @torchensemble_model_doc(
        """Set the training criterion for SoftGradientBoostingClassifier.""",
        "set_criterion",
    )
    def set_criterion(self, criterion):
        super().set_criterion(criterion)

    @_soft_gradient_boosting_model_doc(
        """Implementation on the training stage of SoftGradientBoostingClassifier.""",  # noqa: E501
        "fit",
    )
    def fit(
        self,
        train_loader,
        epochs=100,
        use_reduction_sum=True,
        log_interval=100,
        test_loader=None,
        save_model=True,
        save_dir=None,
    ):
        super().fit(
            train_loader=train_loader,
            epochs=epochs,
            use_reduction_sum=use_reduction_sum,
            log_interval=log_interval,
            test_loader=test_loader,
            save_model=save_model,
            save_dir=save_dir,
        )

    @torchensemble_model_doc(
        """Implementation on the data forwarding in SoftGradientBoostingClassifier.""",  # noqa: E501
        "classifier_forward",
    )
    def forward(self, *x):
        output = [estimator(*x) for estimator in self.estimators_]
        output = op.sum_with_multiplicative(output, self.shrinkage_rate)
        proba = F.softmax(output, dim=1)

        return proba

    @torchensemble_model_doc(item="classifier_evaluate")
    def evaluate(self, test_loader, return_loss=False):
        return super().evaluate(test_loader, return_loss)

    @torchensemble_model_doc(item="predict")
    def predict(self, *x):
        return super().predict(*x)


@_soft_gradient_boosting_model_doc(
    """Implementation on the SoftGradientBoostingRegressor.""", "model"
)
class SoftGradientBoostingRegressor(_BaseSoftGradientBoosting, BaseRegressor):
    def __init__(
        self,
        estimator,
        n_estimators,
        estimator_args=None,
        shrinkage_rate=1.0,
        cuda=True,
        n_jobs=None,
    ):
        super().__init__(
            estimator=estimator,
            n_estimators=n_estimators,
            estimator_args=estimator_args,
            shrinkage_rate=shrinkage_rate,
            cuda=cuda,
            n_jobs=n_jobs,
        )
        self._criterion = nn.MSELoss()
        self.is_classification = False
        self.best_mse = float("inf")

    @torch.no_grad()
    def _evaluate_during_fit(self, test_loader, epoch):
        self.eval()
        mse = 0.0
        flag = False
        criterion = nn.MSELoss()
        for _, elem in enumerate(test_loader):
            data, target = io.split_data_target(elem, self.device)
            output = self.forward(*data)
            mse += criterion(output, target)
        mse /= len(test_loader)

        if mse < self.best_mse:
            self.best_mse = mse
            flag = True

        msg = (
            "Epoch: {:03d} | Validation MSE: {:.5f} | Historical Best: {:.5f}"
        )
        self.logger.info(msg.format(epoch, mse, self.best_mse))
        if self.tb_logger:
            self.tb_logger.add_scalar(
                "soft_gradient_boosting/Validation_MSE", mse, epoch
            )

        return flag

    @torchensemble_model_doc(
        """Set the attributes on optimizer for SoftGradientBoostingRegressor.""",  # noqa: E501
        "set_optimizer",
    )
    def set_optimizer(self, optimizer_name, **kwargs):
        super().set_optimizer(optimizer_name, **kwargs)

    @torchensemble_model_doc(
        """Set the attributes on scheduler for SoftGradientBoostingRegressor.""",  # noqa: E501
        "set_scheduler",
    )
    def set_scheduler(self, scheduler_name, **kwargs):
        super().set_scheduler(scheduler_name, **kwargs)

    @torchensemble_model_doc(
        """Set the training criterion for SoftGradientBoostingRegressor.""",
        "set_criterion",
    )
    def set_criterion(self, criterion):
        super().set_criterion(criterion)

    @_soft_gradient_boosting_model_doc(
        """Implementation on the training stage of SoftGradientBoostingRegressor.""",  # noqa: E501
        "fit",
    )
    def fit(
        self,
        train_loader,
        epochs=100,
        use_reduction_sum=True,
        log_interval=100,
        test_loader=None,
        save_model=True,
        save_dir=None,
    ):
        super().fit(
            train_loader=train_loader,
            epochs=epochs,
            use_reduction_sum=use_reduction_sum,
            log_interval=log_interval,
            test_loader=test_loader,
            save_model=save_model,
            save_dir=save_dir,
        )

    @torchensemble_model_doc(
        """Implementation on the data forwarding in SoftGradientBoostingRegressor.""",  # noqa: E501
        "regressor_forward",
    )
    def forward(self, *x):
        outputs = [estimator(*x) for estimator in self.estimators_]
        pred = op.sum_with_multiplicative(outputs, self.shrinkage_rate)

        return pred

    @torchensemble_model_doc(item="regressor_evaluate")
    def evaluate(self, test_loader):
        return super().evaluate(test_loader)

    @torchensemble_model_doc(item="predict")
    def predict(self, *x):
        return super().predict(*x)



import time
from ctypes import windll

import gym
import pybullet as p
import pybullet_data as pd
from PIL import Image
from numpy.random import Generator

from rrt_ml.utilities.configs import *
from rrt_ml.utilities.formulas import *
from rrt_ml.utilities.hints import *
from rrt_ml.utilities.maps import *
from rrt_ml.utilities.misc import *
from rrt_ml.utilities.paths import *
from rrt_ml.utilities.stats import *


class CarNavigationBulletEnv(gym.Env):

    def __init__(self, cfg: MasterConfig):
        """
        Initialize env.
        """

        self.cfg = cfg  # type: MasterConfig | None
        self.reward_weights = None  # type: Vector2 | None
        self.action_repeat = None  # type: int | None
        self.car = None  # type: Car | None
        self.target = None  # type: Target | None
        self.obstacles = None  # type: list[Obstacle, ...] | None
        self.init_pose = None  # type: Vector3 | None
        self.final_pose = None  # type: Vector3 | None
        self.timestep = None  # type: int | None
        self.bullet_timestep = None  # type: int | None
        self.episode_num = None  # type: int | None
        self.l_ghost_ids = None  # type: list[list[int, ...], ...] | None
        self.save_state = None  # type: int | None
        self.stats = None  # type: EnvStats | None
        self.rng = None  # type: Generator | None
        self.img_crop = None  # type: Vector4 | None
        self.img_extent = None  # type: Vector4 | None

        self._setup()

    def compute_reward(self, achieved_goal: np.ndarray, desired_goal: np.ndarray, info: dict = None):
        """
        Common API for calculating rewards in goal conditioned tasks.
        :param achieved_goal: goal achieved
        :param desired_goal: behavioural goal
        :param info: additional info
        :return: calculated reward
        """

        self._setup_ignore_static_warnings(info)

        # Calculate p-norm reward (works for vectorized calculations)
        abs_dif = np.abs(achieved_goal - desired_goal)
        weighted_abs_dif = np.dot(abs_dif, self.reward_weights)
        reward = -np.power(weighted_abs_dif, self.cfg.env.reward.p_val)

        # Make reward sparse (0 or 1 only)
        reward = self._get_reward_sparse(reward)

        return reward

    def render(self, mode="human") -> None:
        """
        Common API.
        """

        self._setup_ignore_static_warnings()
        print(f'Create env with "GUI" set to "True". Mode: {mode}')

    def reset(self, **kwargs):
        """
        Reset env.
        :param kwargs: k/v 'reset_seed'/bool to set the seed
        :return: initial observation
        """

        self._set_restore_state()
        self._set_place()
        self._set_reset_joints()
        self._set_increment_episode_num()
        self._set_zero_timesteps()
        self._set_stats()

        return self._get_observation()

    def reset_master(self):
        """
        Reset for eval env to track trajectories with deterministic resetting.
        """

        self._set_zero_episode_num()
        self._set_seed()

    def step(self, action: Vector2) -> tuple[dict, float, bool, dict]:
        """
        Step environment.
        :param action: agent action
        :return: tuple [next_obs, reward, done, info]
        """

        # Get current observation to save
        obs = self._get_observation()

        # Scale action from [-1, 1] to [v/phi_min v/phi_max] and set motors refs
        v_ref = scale_to_range(action[0], [-1, 1], [-self.cfg.env.car.v_max, +self.cfg.env.car.v_max])
        phi_ref = scale_to_range(action[1], [-1, 1], [-self.cfg.env.car.phi_max, +self.cfg.env.car.phi_max])
        self.car.set_motors_refs_ackermann(v_ref, phi_ref)

        # Action according to frequency
        for i in range(self.action_repeat):

            # Record stats
            if self.cfg.env.general.stats:

                # Normal stats
                self._set_bullet_stats()
                self._set_bullet_timestep()

            # Step bullet
            p.stepSimulation(physicsClientId=self.client)

        # Calculate reward
        next_obs = self._get_observation()
        reward = self.compute_reward(next_obs['achieved_goal'], next_obs['desired_goal'], {})
        done, done_info = self._get_done_indicator(reward)

        # Update if in eval mode
        if self.cfg.env.general.stats:
            # Update mdp stats
            self._set_mdp_stats(Transition.new(obs, next_obs, action, reward, done))

            # Get top view if done
            # self._set_new_image_top_view()

        # Update timestep
        self._set_increment_mdp_timestep()

        return next_obs, reward, done, {'stats': self.stats, 'done_info': done_info}

    def seed(self, num):
        """
        Do nothing.
        """
        
        return

    def _get_collision_indicator(self):
        """
        Check if the car collided with any obstacle.
        :return: boolean indicator of collision
        """

        # Return true on the first contact
        if self.obstacles is not None:
            for obstacle in self.obstacles:
                contact = (
                        len(
                            p.getContactPoints(
                                self.car.id, obstacle.id, physicsClientId=self.client
                            )
                        )
                        > 0
                )
                if contact:
                    return True

        return False

    def _get_done_indicator(self, reward: float) -> tuple[bool, dict[str, bool]]:
        """
        Check if episode should end.
        :return: boolean indicator and a dict with reasons to terminate (whether they eval as True or False)
        """

        # Get all boolean indicators
        collided = self._get_collision_indicator()
        time_is_up = self._get_time_is_up_indicator()
        goal_reached = self._get_goal_reached_indicator(reward)

        # Get info
        d = dict(collision=collided, time=time_is_up, success=goal_reached)

        return collided or time_is_up or goal_reached, d

    def _get_goal_reached_indicator(self, reward: float) -> bool:
        """
        Check if goal is reached by checking if 'reward' > 'epsilon' (tolerance).
        :param reward: reward at this time step
        :return: boolean indicating if goal was reached
        """

        return reward > self.cfg.env.reward.epsilon

    def _get_image_current_top_view(self) -> Image:
        """
        Get top view image from bullet renderer.
        :return: PIL image
        """

        try:
            import win32gui
            import win32ui
        except ImportError:
            return

        # Get image
        hwnd = win32gui.FindWindow('DeviceWin32', None)
        left, top, right, bot = win32gui.GetClientRect(hwnd)
        windll.user32.SetProcessDPIAware()
        w = right - left
        h = bot - top
        hwnd_dc = win32gui.GetWindowDC(hwnd)
        mfc_dc = win32ui.CreateDCFromHandle(hwnd_dc)
        save_dc = mfc_dc.CreateCompatibleDC()
        save_bit_map = win32ui.CreateBitmap()
        save_bit_map.CreateCompatibleBitmap(mfc_dc, w, h)
        save_dc.SelectObject(save_bit_map)
        windll.user32.PrintWindow(hwnd, save_dc.GetSafeHdc(), 2)
        bmp_info = save_bit_map.GetInfo()
        bmp_str = save_bit_map.GetBitmapBits(True)
        img = Image.frombuffer('RGB', (bmp_info['bmWidth'], bmp_info['bmHeight']), bmp_str, 'raw', 'BGRX', 0, 1)

        # Cleanup
        win32gui.DeleteObject(save_bit_map.GetHandle())
        save_dc.DeleteDC()
        mfc_dc.DeleteDC()
        win32gui.ReleaseDC(hwnd, hwnd_dc)

        # Adjust image
        new_img = img.crop(self.img_crop)
        new_img = new_img.transpose(Image.FLIP_TOP_BOTTOM)

        return new_img

    def _get_observation(self):
        """
        Get env observation.
        :return: dict observation
        """

        # Get values
        state = self.car.get_state()
        achieved_goal = self.car.get_achieved_goal()
        desired_goal = self.target.get_desired_goal()
        mpc_state = self.car.get_state_mpc()

        return dict(
            observation=state, achieved_goal=achieved_goal, desired_goal=desired_goal, mpc_state=mpc_state
        )

    def _get_reward_sparse(self, reward: Vector):
        """
        Make reward sparse (0 or -1 values only).
        :param reward: single reward or array
        :return: sparse reward
        """

        # Vectorized calculation
        if isinstance(reward, list | tuple | np.ndarray):
            for i, r in enumerate(reward):
                if r < self.cfg.env.reward.epsilon:
                    reward[i] = -1
                else:
                    reward[i] = 0
                    # reward[i] = r

        # Single reward
        else:
            if reward < self.cfg.env.reward.epsilon:
                reward = -1
            else:
                reward = 0
                # reward = reward

        return reward

    def _get_time_is_up_indicator(self):
        """
        Check end of episode by number of timesteps.
        :return: boolean indicating that maximum timesteps was reached
        """

        return self.timestep >= (self.cfg.env.general.max_timestep - 1)

    def _set_bullet_stats(self):
        """
        Add bullet physics statistics after each bullet step call.
        """

        # Alias
        ep = self.episode_num
        t = self.bullet_timestep

        # Add basic info
        self.stats.bullet.time[ep, t, ...] = t * 1 / 240
        self.stats.bullet.car_poses[ep, t, ...] = self.car.get_pose()
        self.stats.bullet.target_poses[ep, t, ...] = self.target.get_pose()

        # Calculate reward and add 'done'
        obs = self._get_observation()
        reward = self.compute_reward(obs['achieved_goal'], obs['desired_goal'], {})
        self.stats.bullet.rewards[ep, t, ...] = reward
        self.stats.bullet.dones[ep, t, ...] = self._get_done_indicator(reward)

        # Add steer positions
        steer_pos1, _, _, _ = p.getJointState(self.car.id, 4, physicsClientId=self.client)
        steer_pos2, _, _, _ = p.getJointState(self.car.id, 6, physicsClientId=self.client)
        self.stats.bullet.car_steers[ep, t, ...] = [steer_pos1, steer_pos2]

        # Check velocity magnitude and then signal
        v, _ = p.getBaseVelocity(self.car.id, physicsClientId=self.client)
        pos, vel, _, _ = p.getJointState(self.car.id, 2, physicsClientId=self.client)
        sign = 1 if vel > 0 else -1
        self.stats.bullet.car_velocities[ep, t, ...] = sign * np.linalg.norm(v[:2])

        # Add wall time
        self.stats.bullet.wall_time[ep, t, ...] = time.time()

    def _set_bullet_timestep(self):
        """
        Increment bullet timestep counter.
        """

        self.bullet_timestep += 1

    def _set_place_trajectory_ghosts(self, episode_num: int, stats: EnvStats | None = None):
        """
        Place ghosts in trajectory to take screenshot.
        :param episode_num: number of the eval episode
        :param stats: current env stats or other stats object
        """

        # Get the stats where poses come from
        if stats is None:
            stats = self.stats

        # Indexes = [start, end + step, step]
        interval = self.cfg.env.test.general.ghost_interval
        n_timesteps = len(stats.bullet.car_poses[0, :, 0])
        idxs = np.arange(0, n_timesteps, interval)

        # Get poses and steering angles
        poses = stats.bullet.car_poses[episode_num, idxs, ...]
        steers = stats.bullet.car_steers[episode_num, idxs, ...]

        # Remove zeros
        poses = remove_trailing_zeros(poses).reshape(-1, 3)
        steers = remove_trailing_zeros(steers).reshape(-1, 2)

        self.l_ghost_ids = []
        for pose, steer in zip(poses, steers):
            self._set_place_ghost_car(pose, steer)

    def _set_remove_ghosts(self):
        """
        Remove trajectory ghosts from GUI.
        """

        # Return if no ghosts on scene
        if self.l_ghost_ids is None:
            return
        if isinstance(self.l_ghost_ids, list):
            if len(self.l_ghost_ids) == 0:
                return

        # Remove bodies
        for idd in self.l_ghost_ids:
            p.removeBody(idd, physicsClientId=self.client)

    def _set_increment_episode_num(self):
        """
        Increment number of episodes.
        """

        self.episode_num += 1

    def _set_increment_mdp_timestep(self):
        """
        Set simulation timestep value or increase current value by one.
        """

        self.timestep += 1 if self.timestep is not None else 0

    def _set_mdp_stats(self, transition: 'Transition'):
        """
        Add current time step info if on test mode.
        """

        # Alias
        ep = self.episode_num
        t = self.timestep

        # Add
        self.stats.mdp.states[ep, t, ...] = transition.state
        self.stats.mdp.achieved_goals[ep, t, ...] = transition.achieved_goal
        self.stats.mdp.desired_goals[ep, t, ...] = transition.desired_goal
        self.stats.mdp.next_states[ep, t, ...] = transition.next_state
        self.stats.mdp.next_achieved_goals[ep, t, ...] = transition.next_achieved_goal
        self.stats.mdp.next_desired_goals[ep, t, ...] = transition.next_desired_goal
        self.stats.mdp.actions[ep, t, ...] = transition.action
        self.stats.mdp.rewards[ep, t, ...] = transition.reward
        self.stats.mdp.dones[ep, t, ...] = transition.done

    def _set_place(self):
        """
        Place dynamic objects (car and target)
        """

        # Place car
        car_pose = self.rng.uniform(self.car_reset_low, self.car_reset_high)
        self.car.set_pose(car_pose)

        # Place target
        target_pose = self.rng.uniform(self.target_reset_low, self.target_reset_high)
        self.target.set_pose(target_pose)

    def _set_place_ghost_car(self, pose: Vector3 | None = None, steers: Vector2 | None = None):
        """
        Place transparent visual shape of the car, at its current pose, to log trajectory.
        """

        # Call with argument to set at desired pose
        if pose is not None:
            pos = [pose[0], pose[1], 0.001]
            orn = p.getQuaternionFromEuler([0, 0, pose[2]])
            pos1, pos2 = steers

        # Get car's pose and steering wheels angles
        else:
            pos, orn = p.getBasePositionAndOrientation(self.car.id, physicsClientId=self.client)
            pos1, _, _, _ = p.getJointState(self.car.id, 4, physicsClientId=self.client)
            pos2, _, _, _ = p.getJointState(self.car.id, 6, physicsClientId=self.client)

        # Load URDF at car's current pose
        idd = p.loadURDF('ghost.urdf', pos, orn, useFixedBase=1, physicsClientId=self.client)

        # Change steering angles for the shape
        p.resetJointState(idd, 4, pos1, physicsClientId=self.client)
        p.resetJointState(idd, 6, pos2, physicsClientId=self.client)

        # Initialize list of ghosts if not initialized
        if self.l_ghost_ids is None:
            self.l_ghost_ids = []

        self.l_ghost_ids.append(idd)
        self.l_ghost_ids.append(idd)

    def _set_reset_joints(self):
        """
        Reset car joints to make velocity and steering null.
        """

        joints = [2, 3, 4, 5, 6, 7]
        for joint in joints:
            p.resetJointState(self.car.id, joint, 0, physicsClientId=self.client)

    def _set_seed(self):
        """
        Set seed to make resets deterministic.
        """

        if self.cfg.env.general.seed is not None:
            self.rng = np.random.default_rng(self.cfg.env.general.seed)
        else:
            self.rng = np.random.default_rng()

    def _set_stats(self):
        """
        Dictionary to hold statistics.
        """

        # Reset stats if current episode number is 1
        if self.episode_num == 0:
            # noinspection PyTypeChecker
            self.stats = EnvStats.new(self.cfg)

    def _set_wait_objects_fall(self):
        """
        Wait for placed objects to fall.
        """

        for _ in range(20):
            p.stepSimulation(physicsClientId=self.client)

    def _set_zero_episode_num(self):
        """
        Zero out (actually set it to '-1') episode counter. Call it after setting the seed.
        """

        self.episode_num = -1

    def _set_zero_timesteps(self):
        """
        Reset MDP timesteps counter.
        """

        self.timestep = 0
        self.bullet_timestep = 0

    def _setup(self) -> None:
        """
        Several initializations and configurations.
        """

        self._setup_seed()
        self._setup_client()
        self._setup_action_space()
        self._setup_observation_space()
        self._setup_action_repeat()
        self._setup_reset_bounds()
        self._setup_reward_weights()
        self._setup_urdf_paths()
        self._setup_ground_plane()
        self._setup_gravity()
        self._setup_car()
        self._setup_target()
        self._setup_obstacles()
        self._setup_save_state()
        self._setup_init_final_pose()
        self._setup_camera()
        self._setup_episode_num()

    def _setup_action_repeat(self):
        """
        Set constant for repeating actions according to frequency.
        """

        self.action_repeat = int(1 / (self.cfg.env.car.f_val * (1 / 240)))

    def _setup_action_space(self) -> None:
        """
        Set up gym action space.
        """

        self.action_space = gym.spaces.Box(low=-1.0, high=1.0, shape=(2,), dtype=np.float32)

    def _setup_camera(self):
        """
        Set up matrices for image creation.
        """

        # Get camera distance
        if self.cfg.maps.general.map_name is None:
            self.img_crop = self.cfg.env.test.none.img_crop
            self.img_extent = self.cfg.env.test.none.img_extent
            cam_dist = self.cfg.env.test.none.cam_dist
            cam_target = self.cfg.env.test.none.cam_target

        elif self.cfg.maps.general.map_name == 'narrow':
            self.img_crop = self.cfg.env.test.narrow.img_crop
            self.img_extent = self.cfg.env.test.narrow.img_extent
            cam_dist = self.cfg.env.test.narrow.cam_dist
            cam_target = self.cfg.env.test.narrow.cam_target

        else:
            raise NotImplementedError

        # Remove GUI elements and set top view camera config
        p.configureDebugVisualizer(p.COV_ENABLE_GUI, 0, physicsClientId=self.client)
        p.configureDebugVisualizer(p.COV_ENABLE_SHADOWS, 0, physicsClientId=self.client)
        p.resetDebugVisualizerCamera(cam_dist, 0, -89.99, cam_target, physicsClientId=self.client)

    def _setup_car(self) -> None:
        """
        Initialize car.
        """

        self.car = Car(self.cfg, self.client)

    def _setup_client(self) -> None:
        """
        Assign a client ID, necessary for vectorized environments.
        """

        self.client = p.connect(p.GUI) if self.cfg.env.general.gui else p.connect(p.DIRECT)

    def _setup_episode_num(self) -> None:
        """
        Initialize episode number with '-1' because calling 'reset' should increment it by one.
        """

        self.episode_num = -1

    def _setup_gravity(self):
        """
        Set the gravity.
        """

        p.setGravity(0, 0, -9.81, physicsClientId=self.client)

    def _setup_ground_plane(self) -> None:
        """
        Load the ground plane.
        """

        p.setAdditionalSearchPath(str(Paths().deps_models))
        self.plane = p.loadURDF("meshes/plane.urdf", useFixedBase=1, physicsClientId=self.client)

    def _setup_ignore_static_warnings(self, info=None):
        """
        Stop linting 'method may be static' or unused parameter by putting this method inside another method.
        """

        pass

    def _setup_init_final_pose(self):
        """
        Set attributes 'init/final_pose' as the initial achieved/desired goal.
        """

        self.init_pose = self.car.get_pose()
        self.final_pose = self.target.get_pose()

    def _setup_observation_space(self) -> None:
        """
        Set up the gym observation space. State is [x y sinOrn cosOrn v steer]. Goal is [x y sinOrn cosOrn].
        """

        self.observation_space = gym.spaces.Dict(
            dict(
                observation=gym.spaces.Box(
                    low=-np.inf, high=+np.inf, shape=(6,)
                ),
                desired_goal=gym.spaces.Box(
                    low=-np.inf, high=+np.inf, shape=(4,)
                ),
                achieved_goal=gym.spaces.Box(
                    low=-np.inf, high=+np.inf, shape=(4,)
                ),
            )
        )

    def _setup_obstacles(self) -> None:
        """
        Load all obstacles.
        """

        # Return if there is no obstacles to set up
        if self.cfg.maps.general.map_name is None:
            return

        # Create map and add to obstacles descriptions lists
        m = Map(self.cfg)

        # Iterate over vertices and orientations creating 'Obstacles'
        self.obstacles = []
        for vertex, orn in zip(m.vertices, m.orientations):
            self.obstacles.append(Obstacle(self.cfg, vertex, orn, self.client))

    def _setup_reset_bounds(self) -> None:
        """
        Set up reset bounds.
        """

        # Aliases
        cp = self.cfg.env.car.pose
        tp = self.cfg.env.target.pose

        # Set car pose
        if isinstance(cp, float | int):
            self.car_reset_low = [-cp, -cp, 0]
            self.car_reset_high = [+cp, +cp, 2 * np.pi]
        elif isinstance(cp, list | tuple | np.ndarray):
            self.car_reset_low = [cp[0], cp[1], cp[2]]
            self.car_reset_high = [cp[0], cp[1], cp[2]]

        # Set target pose
        if isinstance(tp, float | int):
            self.target_reset_low = [-tp, -tp, 0]
            self.target_reset_high = [+tp, +tp, 2 * np.pi]
        elif isinstance(tp, list | tuple | np.ndarray):
            self.target_reset_low = [tp[0], tp[1], tp[2]]
            self.target_reset_high = [tp[0], tp[1], tp[2]]

    def _set_restore_state(self):
        """
        Restore state to maintain contact information.
        """

        p.restoreState(self.save_state, physicsClientId=self.client)

    def _setup_reward_weights(self) -> None:
        """
        Set up the reward weights for each state component.
        """

        self.reward_weights = 0.5 * np.array(
            [
                self.cfg.env.reward.weights[0],
                self.cfg.env.reward.weights[0],
                self.cfg.env.reward.weights[1],
                self.cfg.env.reward.weights[1],
            ]
        )

    def _setup_save_state(self):
        """
        Save all information to guarantee deterministic resetting.
        """

        # Get contact information for car
        self.car.set_pose([0, 0, 0])
        self.target.set_pose([0, 0, 0])
        for _ in range(10):
            p.stepSimulation(physicsClientId=self.client)

        # Save this state
        self.save_state = p.saveState(physicsClientId=self.client)

    def _setup_seed(self):
        """
        Set up random seed for placing car and target. Deterministic only if setting the seed through other method.
        """

        self.rng = np.random.default_rng()

    def _setup_target(self) -> None:
        """
        Initialize target.
        """

        self.target = Target(self.cfg, client=self.client)

    def _setup_urdf_paths(self) -> None:
        """
        Set additional search paths for pybullet.
        """

        p.setAdditionalSearchPath(pd.getDataPath(), physicsClientId=self.client)


class Car:

    def __init__(self, cfg: MasterConfig, client: int = 0):
        """
        Initialize car.
        :param client: pybullet client id
        :param cfg: config object
        """

        self.client = client
        self.cfg = cfg

        self.id = None  # type: None | int

        self._setup()

    def get_achieved_goal(self):
        """
        Get car achieved goal.
        :return: vector [x y sinOrn cosOrn]
        """

        # Get x, y and theta
        pose = self.get_pose()
        state = pose3_to_state4(pose)

        return state

    def get_pose(self):
        """
        Get car position and orientation.
        :return: car pose
        """

        # Get from PyBullet
        pos, orn = p.getLinkState(self.id, 0, physicsClientId=self.client)[:2]

        # Get orientation
        theta = quaternion_to_theta(orn)

        return np.array([pos[0], pos[1], theta])

    def get_state(self):
        """
        Get car state.
        :return: vector [x y sinOrn cosOrn v phi]
        """

        # Get x, y and theta
        pose = self.get_pose()
        state = pose3_to_state4(pose)

        # Get base velocity
        lin_vel = p.getBaseVelocity(self.id, physicsClientId=self.client)
        v = np.linalg.norm([lin_vel[0], lin_vel[1]])

        # Get steering angle
        phi = p.getJointState(self.id, 4, physicsClientId=self.client)[0]

        return np.array([*state, v, phi])

    def get_state_mpc(self):
        """
        Get state for MPC controller (velocity is signed).
        :return: state as [x y theta v phi]
        """

        # Get pose
        pose = self.get_pose()

        # Get velocity with sign
        lin_vel = p.getBaseVelocity(self.id, physicsClientId=self.client)
        vel = np.linalg.norm([lin_vel[0], lin_vel[1]])
        joint_pos, joint_vel, _, _ = p.getJointState(self.id, 2, physicsClientId=self.client)
        sign = 1 if joint_vel > 0 else -1
        vel = sign * vel

        # Get steering angle
        a = self.cfg.env.car.axis_dist
        b = self.cfg.env.car.wheel_dist
        phir, _, _, _ = p.getJointState(self.id, 6)
        phi = np.arctan((2 * a * np.tan(phir)) / (2 * a - b * np.tan(phir)))

        return np.array([*pose, vel, phi])

    def set_pose(self, pose: Vector3):
        """
        Set the target to a certain pose.
        :param pose: vector of 3 components: (x y angle[rad])
        """

        # Use PyBullet API
        x, y = pose[0], pose[1]
        z = self.cfg.env.car.reset_z
        orn_quaternion = p.getQuaternionFromEuler([0, 0, pose[2]])
        p.resetBasePositionAndOrientation(self.id, [x, y, z], orn_quaternion, physicsClientId=self.client)

        # Perform collision detection because we may place the car at different poses to check free regions
        p.performCollisionDetection(physicsClientId=self.client)

    def set_motors_refs_ackermann(self, v_ref: float, phi_ref: float) -> None:
        """
        Move car according to the Ackermann's geometry.
        :param v_ref: desired linear velocity
        :param phi_ref: desired steering angle
        """

        # Use formula to calculate velocities and steering angles
        vrl, vrr, vfl, vfr, phil, phir = get_ackermann_v_rf_lr_phi_lr(v_ref, phi_ref, self.cfg)

        # Set linear velocities
        v_dot_max = self.cfg.env.car.v_dot_max
        p.setJointMotorControl2(self.id, 7, p.VELOCITY_CONTROL, force=v_dot_max, targetVelocity=vfr,
                                physicsClientId=self.client)
        p.setJointMotorControl2(self.id, 5, p.VELOCITY_CONTROL, force=v_dot_max, targetVelocity=vfl,
                                physicsClientId=self.client)
        p.setJointMotorControl2(self.id, 3, p.VELOCITY_CONTROL, force=v_dot_max, targetVelocity=vrr,
                                physicsClientId=self.client)
        p.setJointMotorControl2(self.id, 2, p.VELOCITY_CONTROL, force=v_dot_max, targetVelocity=vfl,
                                physicsClientId=self.client)

        # Set steering angles
        phi_dot_max = self.cfg.env.car.phi_dot_max
        p.setJointMotorControl2(self.id, 6, p.POSITION_CONTROL, maxVelocity=phi_dot_max, targetPosition=phir,
                                physicsClientId=self.client)
        p.setJointMotorControl2(self.id, 4, p.POSITION_CONTROL, maxVelocity=phi_dot_max, targetPosition=phil,
                                physicsClientId=self.client)

    def _setup(self):
        """
        Set up.
        """

        self._setup_id()

    def _setup_id(self):
        """
        Set up urdf.
        """

        p.setAdditionalSearchPath(str(Paths().deps_models), physicsClientId=self.client)
        self.id = p.loadURDF("racecar.urdf", physicsClientId=self.client)

    def _setup_ignore_static_warnings(self):
        """
        Ignore PyCharm static method warnings.
        """

        pass


class Target:

    def __init__(self, cfg: MasterConfig, client: int):
        """
        Initialize target as desired pose.
        """

        self.cfg = cfg
        self.client = client

        self.id = None

        self._setup()

    def get_pose(self):
        """
        Get target pose.
        :return: vector [x y sinTheta cosTheta]
        """

        # Get from PyBullet
        pos, orn = p.getBasePositionAndOrientation(self.id, physicsClientId=self.client)

        # Get orientation
        theta = quaternion_to_theta(orn)

        return np.array([pos[0], pos[1], theta])

    def get_desired_goal(self):
        """
        Get desired goal.
        :return: vector [x y sinTheta cosTheta]
        """

        # Get x, y and theta
        pose = self.get_pose()
        state = pose3_to_state4(pose)

        return state

    def set_pose(self, pose: Vector3):
        """
        Set the target to a certain pose.
        :param pose: vector of 3 components: (x y angle[rad])
        """

        # Use PyBullet API
        x, y = pose[0], pose[1]
        z = self.cfg.env.target.reset_z
        orn_quaternion = p.getQuaternionFromEuler([0, 0, pose[2]])
        p.resetBasePositionAndOrientation(self.id, [x, y, z], orn_quaternion, physicsClientId=self.client)

    def _setup(self):
        """
        Set up.
        """

        self._setup_id()

    def _setup_id(self):
        """
        Set up urdf.
        """

        p.setAdditionalSearchPath(str(Paths().deps_models), physicsClientId=self.client)
        self.id = p.loadURDF('target.urdf', useFixedBase=1, globalScaling=0.025, physicsClientId=self.client)


class Obstacle:

    def __init__(self, cfg: MasterConfig, top_left_bottom_right: Vector4, orn: float, client: int):
        """
        Initialize.
        """

        self.cfg = cfg
        self.top_left_bottom_right = top_left_bottom_right
        self.orn = orn
        self.client = client

        self.id = None

        self._setup()

    def _setup(self):
        """
        Set up.
        """

        self._setup_id()

    def _setup_id(self):
        """
        Set up id.
        """

        # We set up a collision and visual shape
        ln = self.top_left_bottom_right[2] - self.top_left_bottom_right[0]
        w = self.top_left_bottom_right[1] - self.top_left_bottom_right[3]
        h = self.cfg.maps.obstacles.height
        c = self.cfg.maps.obstacles.color
        m = self.cfg.maps.obstacles.mass
        z = self.cfg.maps.obstacles.reset_z

        # Create collision and visual shapes
        collision_shape = p.createCollisionShape(
            shapeType=p.GEOM_BOX,
            halfExtents=[ln / 2, w / 2, h],
            physicsClientId=self.client,
        )
        visual_shape = p.createVisualShape(
            shapeType=p.GEOM_BOX,
            halfExtents=[ln / 2, w / 2, h],
            rgbaColor=c,
            physicsClientId=self.client,
        )

        # Create a heavy multi-body, so that it can't be moved by the car
        x = self.top_left_bottom_right[0] + ln / 2
        y = self.top_left_bottom_right[3] + w / 2
        orn = p.getQuaternionFromEuler([0, 0, self.orn])
        self.id = p.createMultiBody(
            baseMass=m,
            baseCollisionShapeIndex=collision_shape,
            baseVisualShapeIndex=visual_shape,
            basePosition=[x, y, z],
            baseOrientation=orn,
            physicsClientId=self.client,
        )


class Transition(BaseStats):
    action: Vector2 | None = None
    reward: float | None = None
    done: bool | None = None
    desired_goal: np.ndarray | None = None
    achieved_goal: np.ndarray | None = None
    state: np.ndarray | None = None
    next_desired_goal: np.ndarray | None = None
    next_achieved_goal: np.ndarray | None = None
    next_state: np.ndarray | None = None

    @classmethod
    def new(cls, obs, next_obs, action, reward, done):
        t = Transition()
        t.desired_goal = obs['desired_goal']
        t.achieved_goal = obs['achieved_goal']
        t.state = obs['observation']
        t.next_desired_goal = next_obs['desired_goal']
        t.next_achieved_goal = next_obs['achieved_goal']
        t.next_state = next_obs['observation']
        t.action = action
        t.reward = reward
        t.done = done
        return t



import math
import time
from itertools import product
from pathlib import Path

import numpy as np
import pandas as pd
import pybullet as p
import seaborn as sns
import yaml
from matplotlib import pyplot as plt

from rrt_ml.algorithms.rl import *
from rrt_ml.algorithms.rrt import *
from rrt_ml.algorithms.sl import *
from rrt_ml.utilities.analytic import *
from rrt_ml.utilities.configs import *
from rrt_ml.utilities.maps import *
from rrt_ml.utilities.paths import *


class Hyper:

    def __init__(self, cfg: MasterConfig):
        """
        Initialize.
        :param cfg: general config
        """

        self.cfg = cfg
        self.cfgs = None  # type: None | list[MasterConfig, ...]
        self.algorithm = None  # type: RL | SL | RRT | None
        self.path = None  # type: None | Path
        self.prefix = None  # type: None | str
        self.hyper_dict = None  # type: None | dict

        self._setup()

    def train(self):
        """
        Train all hyperparameters.
        """

        # Test for different configs
        for i, cfg in enumerate(self.cfgs):

            # Don't load if next config experiment folder exists
            # if i != len(self.cfgs) - 1:
            #     next_cfg_name = self.cfgs[i+1].general.config_name_or_prefix
            #     next_cfg_exp_path = Paths().exp('rrt', next_cfg_name)
            #     if next_cfg_exp_path.exists():
            #         continue

            # Train
            alg = self.algorithm(cfg)
            alg.train()

    def test(self):
        """
        Get various results.
        """

        match self.cfg.general.algorithm:
            case 'rl':
                self._test_rl()
            case 'rrt':
                self._test_rrt()
            case 'sl':
                self._test_sl()
            case _:
                raise NotImplementedError

    def _test_rl(self):
        """
        Test RL configs.
        """

        self._test_rl_plot_rewards_entropy()

    def _test_rl_plot_rewards_entropy(self):
        """
        Plot rewards for all configs
        """

        # Dict to hold data
        l_stats = []
        l_names = []

        # Loop configs
        for cfg in self.cfgs:

            # Load RL, get stats and name
            l_stats.append(RL(cfg).stats)
            l_names.append(cfg.general.config_name_or_prefix)

        # Figures
        fig, axs = plt.subplots(2, 2, sharey='row', sharex='col', figsize=(9, 6))
        fig.subplots_adjust(wspace=0.15, hspace=0.2)
        axs = axs.flatten()

        for i, ax in enumerate(axs):

            # Plot reward
            ax.plot(l_stats[i].train_timestep_nums, l_stats[i].total_val_rewards, label='Recompensa', color='blue')

            # Set tick color to blue
            ax.tick_params(axis='y', labelcolor='blue')

            # Get entropy
            csv_path = next(Paths().exp('rl', l_names[0]).glob('*entropy*'))
            df = pd.read_csv(str(csv_path))

            # Plot
            axt = ax.twinx()
            axt.plot(df['step'][7:], df['Explore/ag_kde_entropy'][7:], label='Entropia', color='green')

            # Set color to green
            axt.tick_params(axis='y', labelcolor='green')

            # Legends
            ax.legend(loc='lower right')
            axt.legend(bbox_to_anchor=(1, 0.31))

            # Grid
            ax.grid(True)

            # X axis label
            if i in [2, 3]:
                ax.set_xlabel('Passos de treinamento')

        fig_name = f'rlHiperRecompensaEntropia.png'
        fig.savefig(str(self.path / fig_name), dpi=600, bbox_inches='tight')

    def _test_rrt(self):
        """
        Test RRT configs.
        """

        # Generate results for CVAE experiment
        if 'cvae_rs' in self.cfg.general.config_name_or_prefix:
            self._test_rrt_compare_time_to_solve_cvae()

        # Generate all other results
        df = pd.DataFrame({
            'Critério': [],
            'Custo': [],
            'Tempo de resolução (s)': [],
            'Tempo de inferência': [],
            'Sucesso': [],
            'Semente': [],
            'RRT*': []
        })

        # Loop over all configs to generate test results
        loaded_df = False
        for cfg in self.cfgs:

            # Break if dataframe exists
            if (self.path / 'results_dataframe.csv').exists():
                df = pd.read_csv(str(self.path / 'results_dataframe.csv'))
                loaded_df = True
                break

            # Test function will generate simulation info attribute and figures
            rrt = RRT(cfg)
            rrt.test()

            # Decide what is the RRT* variation
            match cfg.rrt.general.controller:
                case 'rl':
                    rrt_var = 'ApM'
                case _:
                    rrt_var = 'RS+MPC'

            # Add row for both criterion
            time_to_solve = rrt._test_get_time_to_solve()
            df_row1 = pd.DataFrame({
                'Critério': ['Tempo (s)'],
                'Custo': [rrt._test_get_solution_simulation_cost_time()],
                'Tempo de resolução (s)': [time_to_solve],
                'Tempo de inferência': [rrt.solution_simulation_info.cpu_time],
                'Sucesso': [rrt.solution_simulation_info.success],
                'Semente': [cfg.rrt.general.seed],
                'RRT*': [rrt_var],
            })
            df_row2 = pd.DataFrame({
                'Critério': ['Distância (m)'],
                'Custo': [rrt._test_get_solution_simulation_cost_length()],
                'Tempo de resolução (s)': [time_to_solve],
                'Tempo de inferência': [rrt.solution_simulation_info.cpu_time],
                'Sucesso': [rrt.solution_simulation_info.success],
                'Semente': [cfg.rrt.general.seed],
                'RRT*': [rrt_var],
            })

            # Add row to main dataframe
            df = pd.concat((df, df_row1, df_row2), ignore_index=True)

            # Top view of solution
            self._test_rrt_solution_top_view(cfg, rrt)

            # Tree of solution
            self._test_rrt_solution_tree(cfg, rrt)

            # Disconnect gui env
            p.disconnect(rrt.env.client)

        # Save dataframe
        if not loaded_df:
            # noinspection PyTypeChecker
            df.to_csv(str(self.path / 'results_dataframe.csv'), index=0)

        # Comparison of costs
        #self._test_rrt_compare_success_rates(df)
        #self._test_rrt_compare_costs(df)
        #self._test_rrt_compare_resolution_time(df)
        #self._test_rrt_compare_inference_time()
        #self._test_rrt_plot_all_trees()
        self._test_rrt_plot_all_top_view()

    def _test_rrt_compare_costs(self, df):
        """
        Compare solution costs across approaches.
        :param df: data table
        """

        # Compare mean value when success
        df_copy = df.copy(deep=True)
        seeds = np.unique(df.loc[df['Sucesso'] == 0.0, 'Semente'].values)
        for i, row in df.iterrows():
            if row['Semente'] in seeds:
                df_copy = df_copy.drop(i)

        # Save figure
        path = Paths().exp('rrt', self.prefix) / 'comparacaoCustos.png'
        sns.set_style('whitegrid')
        ax = sns.barplot(data=df_copy, x="Critério", y="Custo", hue="RRT*")
        ax.get_figure().savefig(str(path), dpi=600, bbox_inches='tight')
        plt.close()

    def _test_rrt_compare_inference_time(self):
        """
        Compare inference time for controllers.
        """

        # Start
        rrt = RRT(self.cfgs[0])
        obs = rrt.env.reset()
        rrt._get_action_rl(obs)
        mpcrs = MPCRS(self.cfgs[0])
        mpcrs.get_action(obs)

        # Check time
        a = time.time()
        for _ in range(100):
            rrt._get_action_rl(obs)
        time_rl = (time.time() - a) / 100 * 1000
        a = time.time()
        for _ in range(100):
            mpcrs.get_action(obs)
        time_mpc = (time.time() - a) / 100 * 1000

        # Get dataframe
        df = pd.DataFrame(data={'Tempo de inferência (ms)': [time_rl, time_mpc],
                                'Controlador': ['Agente', 'CPM']})

        # # Save figure
        path = Paths().exp('rrt', self.prefix) / 'tempoInferenciaControle.png'
        sns.set_style('whitegrid')
        ax = sns.barplot(data=df, x="Tempo de inferência (ms)", y="Controlador", orient='h')
        ax.get_figure().savefig(str(path), dpi=600, bbox_inches='tight')
        plt.close()

    def _test_rrt_compare_resolution_time(self, df):
        """
        Compare time to find a solution in RRT.
        :param df: data table
        """

        # Save figure
        path = Paths().exp('rrt', self.prefix) / 'custoComputacionalRrtEstrela.png'
        sns.set_style('whitegrid')
        ax = sns.barplot(data=df, x="RRT*", y="Tempo de resolução (s)")
        ax.get_figure().savefig(str(path), dpi=600, bbox_inches='tight')
        plt.close()

    def _test_rrt_compare_success_rates(self, df):
        """
        Compare success rate for the controllers.
        :param df: data table
        """

        # Save figure
        path = Paths().exp('rrt', self.prefix) / 'rrtTaxaSucessoControladores.png'
        sns.set_style('whitegrid')
        ax = sns.barplot(x='RRT*', y='Sucesso', data=df, estimator=lambda x: sum(x == 1) * 100.0 / len(x))
        ax.set(ylabel='Taxa de sucesso (%)')
        ax.get_figure().savefig(str(path), dpi=600, bbox_inches='tight')
        plt.close()

    def _test_rrt_compare_time_to_solve_cvae(self):
        """
        Compare resolution time and costs when using cvae.
        """

        path = Paths().exp('rrt', self.prefix) / 'tempoResolucaoAlpha.png'
        if path.exists():
            return

        times_cvae = []
        times_uniform = []
        costs_cvae = []
        costs_uniform = []
        for cfg in self.cfgs:

            # Initialize algo
            rrt = RRT(cfg)

            # Check config
            match cfg.rrt.sample.learner_prob:
                case 0:
                    times_uniform.append(rrt._test_get_time_to_solve())
                    costs_uniform.append(rrt.node_final.cost_from_root)
                case _:
                    times_cvae.append(rrt._test_get_time_to_solve())
                    costs_cvae.append(rrt.node_final.cost_from_root)

            # Disconnect env
            p.disconnect(physicsClientId=rrt.env.client)

        # Get mean values
        time_cvae = np.mean(times_cvae)
        time_uniform = np.mean(times_uniform)
        cost_cvae = np.mean(costs_cvae)
        cost_uniform = np.mean(costs_uniform)

        # Normalize to 0-1
        prop_time = time_cvae / time_uniform
        prop_cost = cost_cvae / cost_uniform

        # Plot with seaborn
        df = pd.DataFrame([
            ['Tempo de resolução (s)', 1, r'$\alpha=0$'],
            ['Tempo de resolução (s)', prop_time, r'$\alpha=0.7$'],
            ['Custo (m)', 1, r'$\alpha=0$'],
            ['Custo (m)', prop_cost, r'$\alpha=0.7$'],
        ], columns=['Critério', 'Valor', r'$\alpha$'])

        sns.set(font_scale=1.2)
        sns.set_style('whitegrid')
        ax = sns.barplot(data=df, x="Critério", y="Valor", hue='$\\alpha$')
        ax.set(yticklabels=[])  # remove the tick labels
        ax.set(ylabel='Valores normalizados')
        sns.move_legend(ax, 'upper center')
        ax.get_figure().savefig(str(path), dpi=600)
        plt.close()

    def _test_rrt_plot_all_top_view(self):
        """
        Plot and compare all trees and solutions for all problems.
        """

        # Get path
        path = Paths().exp('rrt', self.prefix)
        path1 = path / 'topViewTodas1.png'
        path2 = path / 'topViewTodas2.png'
        # if path1.exists() and path2.exists():
        #     return

        # Create figures
        figs, l_axs = [], []
        for i in range(2):
            fig, axs = plt.subplots(3, 3, sharey='row', sharex='col', figsize=(12, 9.6))
            fig.subplots_adjust(wspace=-0.4, hspace=0.3)
            figs.append(fig)
            l_axs.append(axs.flatten())
        all_axs = np.concatenate(l_axs)

        # Loop axes
        for i, (fig, axs) in enumerate(zip(figs, l_axs)):
            for j, ax in enumerate(axs):

                # Initialize algo and map
                rrt = RRT(self.cfgs[i * 9 + j])

                # Remove car and target from origin
                rrt.env.car.set_pose([-10, -10, 0])
                rrt.env.target.set_pose([-10, -10, 0])

                # Place ghosts, get image and then remove ghosts
                rrt.env._set_place_trajectory_ghosts(episode_num=0, stats=rrt.solution_simulation_info.ep_stats)
                img = rrt.env._get_image_current_top_view()
                rrt.env._set_remove_ghosts()
                ax.imshow(img, origin='lower', extent=rrt.cfg.env.test.narrow.img_extent)

                # Check which algorithm is running
                alg = 'CPM' if rrt.cfg.rrt.general.controller == 'mpc' else 'Agente'

                # Axes labels
                if j in [0, 3, 6]:
                    ax.set_ylabel('y [m]')
                if j in [6, 7, 8]:
                    ax.set_xlabel('x [m]')

                # Title
                ax.set_title(f'Problema {j+1} - {alg}')

                # Disconnect test env
                p.disconnect(physicsClientId=rrt.env.client)

        # Save figure
        figs[0].savefig(path1, dpi=600, bbox_inches='tight', pad_inches=0.1)
        figs[1].savefig(path2, dpi=600, bbox_inches='tight', pad_inches=0.1)
        plt.close()

    def _test_rrt_plot_all_trees(self):
        """
        Plot and compare all trees and solutions for all problems.
        """

        # Get path
        path = Paths().exp('rrt', self.prefix)
        path1 = path / 'arvoreSolucaoTodas1.png'
        path2 = path / 'arvoreSolucaoTodas2.png'
        if path1.exists() and path2.exists():
            return

        # Create figures
        figs, l_axs = [], []
        for i in range(2):
            fig, axs = plt.subplots(3, 3, sharey='row', sharex='col', figsize=(12, 9.6))
            fig.subplots_adjust(wspace=-0.4, hspace=0.3)
            figs.append(fig)
            l_axs.append(axs.flatten())
        all_axs = np.concatenate(l_axs)

        # Loop axes
        for i, (fig, axs) in enumerate(zip(figs, l_axs)):
            for j, ax in enumerate(axs):

                # Initialize algo and map
                rrt = RRT(self.cfgs[i*9 + j])
                m = Map(rrt.cfg)

                # Add normal nodes
                m.set_add_nodes(rrt.list_nodes)
                m.get_plot_tree(ax)

                # Add solution
                m._set_plot_branches_for_list(ax, list_nodes=rrt.solution_simulation_info.list_nodes)

                # Add initial and final nodes
                nodes = [rrt.solution_simulation_info.list_nodes[0], rrt.solution_simulation_info.list_nodes[-1]]
                m._set_plot_nodes(ax, nodes=nodes)

                # Axes labels
                if j in [0, 3, 6]:
                    ax.set_ylabel('y [m]')
                if j in [6, 7, 8]:
                    ax.set_xlabel('x [m]')

                # Title
                ax.set_title(f'Problema {j+1}')

                # Disconnect test env
                p.disconnect(physicsClientId=rrt.env.client)

        # Save figure
        figs[0].savefig(path1, dpi=600, bbox_inches='tight', pad_inches=0.1)
        figs[1].savefig(path2, dpi=600, bbox_inches='tight', pad_inches=0.1)
        plt.close()

    def _test_rrt_solution_top_view(self, cfg, rrt):
        """
        Get top view of solution trajectory
        :param cfg: current config
        :param rrt: RRT* instance
        """

        # Get path
        path = Paths().exp('rrt', self.prefix) / ('solucaoTopView' + '_' + cfg.general.config_name_or_prefix + '.png')
        if path.exists():
            return

        # Place ghosts, get image and then remove ghosts
        rrt.env._set_place_trajectory_ghosts(episode_num=0, stats=rrt.solution_simulation_info.ep_stats)
        img = rrt.env._get_image_current_top_view()
        rrt.env._set_remove_ghosts()

        # Plot on axis
        fig, ax = plt.subplots()
        ax.imshow(img, origin='lower', extent=cfg.env.test.narrow.img_extent)
        fig.savefig(path, dpi=600, bbox_inches='tight')
        plt.close()

    def _test_rrt_solution_tree(self, cfg, rrt):
        """
        Get tree view of solution trajectory
        :param cfg: current config
        :param rrt: RRT* instance
        """

        # Get path
        path = Paths().exp('rrt', self.prefix)
        path = path / ('arvoreSolucaoTopView' + '_' + cfg.general.config_name_or_prefix + '.png')
        if path.exists():
            return

        # Make tree
        mmap = Map(cfg)
        mmap.set_add_nodes(rrt.list_nodes)
        fig, ax = mmap.get_plot_tree()

        # Make solution line
        mmap._set_plot_branches_for_list(ax, list_nodes=rrt.solution_simulation_info.list_nodes)

        # Plot initial and final nodes
        nodes = [rrt.solution_simulation_info.list_nodes[0], rrt.solution_simulation_info.list_nodes[-1]]
        mmap._set_plot_nodes(ax, nodes=nodes)

        # Axis info
        ax.set_xlabel('x [m]')
        ax.set_ylabel('y [m]')

        # Save figure
        fig.savefig(path, dpi=600, bbox_inches='tight', pad_inches=0.1)
        plt.close()

    def _test_sl(self):
        """
        Test SL configs.
        """

        # self._test_sl_compare_samples_all()
        # self._test_sl_compare_samples_eight()
        # self._test_sl_compare_samples_two()
        self._test_sl_compare_error_three()

    def _test_sl_compare_samples_all(self):
        """
        Compare samples of all configs.
        """

        # Constants
        n_samples = 15
        seed = 0
        map_nums = [0, 20, 40, 60, 80, 100, 120]

        # Loop over maps
        for map_num in map_nums:

            # Return if figure exists
            name = f'sl_hyper_comparacao_amostras_todos_mapa_{map_num}_fig_0.png'
            if (self.path / name).exists():
                continue

            # Constants
            n_rows = 4
            n_cols = 4
            n_cfgs = len(self.cfgs)
            n_figs = math.ceil(n_cfgs / (n_rows * n_cols))

            # Create figures, flatten list of axes before adding to list and make a big list of axes
            figs, l_axs = [], []
            for i in range(n_figs):
                fig, axs = plt.subplots(n_rows, n_cols, sharey='row', sharex='col', figsize=(12, 9.6))
                fig.subplots_adjust(wspace=-0.4, hspace=0.3)
                figs.append(fig)
                l_axs.append(axs.flatten())
            all_axs = np.concatenate(l_axs)

            # Get conditions
            obstacles, state_i, state_f = SL(self.cfgs[0]).get_conditions(map_num)

            # Loop through configs and axes
            for i, (cfg, ax) in enumerate(zip(self.cfgs, all_axs)):

                # Get states
                states = SL(cfg).get_samples(map_num, n_samples, np.random.default_rng(seed))

                # Update config before drawing map (obstacles plot comes from config)
                cfg.maps.narrow.narrow1_pos = obstacles[0:2]
                cfg.maps.narrow.narrow2_pos = obstacles[2:4]

                # Initialize map and add states
                mapp = Map(cfg)
                mapp.set_add_states(states, 'sl')
                mapp.set_add_states(state_i, 'init')
                mapp.set_add_states(state_f, 'final')

                # Add plot to current axis
                mapp.get_plot_lines(ax)

                # Axes labels
                if i % n_cols == 0:
                    ax.set_ylabel('y [m]')
                if (i % (n_rows * n_cols)) >= (n_cols * n_rows - n_cols):
                    ax.set_xlabel('x [m]')

                # Title
                ax.set_title(cfg.general.config_name_or_prefix)

            # Save figures
            for i, fig in enumerate(figs):
                fig_name = f'sl_hyper_comparacao_amostras_todos_mapa_{map_num}_fig_{i}.png'
                fig.savefig(str(self.path / fig_name), dpi=600, bbox_inches='tight')

    def _test_sl_compare_samples_eight(self):
        """
        Compare two configs on 4 maps.
        """

        # Return if figure exists
        path = self.path / 'sl_hyper_comparacao_amostras_oito.png'
        if path.exists():
            return

        # Constants
        # cfg_names = ['h_2421', 'h_2421', 'h_2421', 'h_2421', 'h_2521', 'h_2521', 'h_2521', 'h_2521']
        # map_nums = [5, 6, 7, 8, 5, 6, 7, 8]
        cfg_names = ['h_2121', 'h_2221', 'h_2321', 'h_2421', 'h_2521', 'h_2621', 'h_2721', 'h_2821']
        map_nums = [5, 5, 5, 5, 5, 5, 5, 5]
        n_samples = 15
        seed = 0

        # Constants
        n_rows = 2
        n_cols = 4

        # Create figure
        fig, axs = plt.subplots(n_rows, n_cols, sharey='row', sharex='col', figsize=(12, 9.6))
        fig.subplots_adjust(hspace=-0.45, wspace=0.15)
        axs = axs.flatten()

        # Loop through configs and axes
        for i, (cfg_name, ax, map_num) in enumerate(zip(cfg_names, axs, map_nums)):

            # Get config and SL
            cfg = self._get_config(cfg_name)
            sl = SL(cfg)

            # Get conditions, states and delete SL
            obstacles, state_i, state_f = sl.get_conditions(map_num)
            states = sl.get_samples(map_num, n_samples, np.random.default_rng(seed))
            del sl

            # Update config before drawing map (obstacles plot comes from config)
            cfg.maps.narrow.narrow1_pos = obstacles[0:2]
            cfg.maps.narrow.narrow2_pos = obstacles[2:4]

            # Initialize map and add states
            mapp = Map(cfg)
            mapp.set_add_states(states, 'sl')
            mapp.set_add_states(state_i, 'init')
            mapp.set_add_states(state_f, 'final')

            # Add plot to current axis
            mapp.get_plot_lines(ax)

            # Axes labels
            if i % n_cols == 0:
                ax.set_ylabel('y [m]')
            if (i % (n_rows * n_cols)) >= (n_cols * n_rows - n_cols):
                ax.set_xlabel('x [m]')

            # Title
            ax.set_title(cfg.general.config_name_or_prefix + ' (' + r'$\beta=$' + str(cfg.sl.loss.beta) + ')', )

        # Save figure
        fig.savefig(str(path), dpi=600, bbox_inches='tight')

    def _test_sl_compare_samples_two(self):
        """
        Compare state generation of two configs.
        """

        # Return if figure exists
        path = self.path / 'sl_hyper_comparacao_amostras_dois.png'
        if path.exists():
            return

        # Constants
        cfg1_name, cfg2_name = ['h_2421', 'h_2222']
        map_num = 1
        n_samples = 15
        seed = 0

        # Get the two configs to compare
        cfg1, cfg2 = None, None
        for cfg in self.cfgs:
            cfg1 = cfg if cfg.general.config_name_or_prefix == cfg1_name else cfg1
            cfg2 = cfg if cfg.general.config_name_or_prefix == cfg2_name else cfg2

        # Get states and conditions
        states1 = SL(cfg1).get_samples(map_num, n_samples, np.random.default_rng(seed))
        states2 = SL(cfg2).get_samples(map_num, n_samples, np.random.default_rng(seed))
        obstacles, state_i, state_f = SL(cfg1).get_conditions(map_num)

        # Update config with obstacles
        cfg1.maps.narrow.narrow1_pos = obstacles[0:2]
        cfg1.maps.narrow.narrow2_pos = obstacles[2:4]
        cfg2.maps.narrow.narrow1_pos = obstacles[0:2]
        cfg2.maps.narrow.narrow2_pos = obstacles[2:4]

        # Get maps
        map1 = Map(cfg1)
        map2 = Map(cfg2)

        # Add samples, initial and final states
        map1.set_add_states(states1, 'sl')
        map1.set_add_states(state_i, 'init')
        map1.set_add_states(state_f, 'final')
        map2.set_add_states(states2, 'sl')
        map2.set_add_states(state_i, 'init')
        map2.set_add_states(state_f, 'final')

        # Plot maps
        fig, axs = plt.subplots(1, 2, sharey='all', squeeze=True, figsize=(9, 6))
        fig.subplots_adjust(hspace=-0.2)
        map1.get_plot_lines(axs[0])
        map2.get_plot_lines(axs[1])

        # Adjust axes
        axs[0].set_ylabel('y [m]')
        axs[0].set_xlabel('x [m]')
        axs[1].set_xlabel('x [m]')

        # Title
        axs[0].set_title(cfg1.general.config_name_or_prefix)
        axs[1].set_title(cfg2.general.config_name_or_prefix)

        # Save figure
        fig.savefig(str(path), dpi=600, bbox_inches='tight')

    def _test_sl_compare_error_three(self):
        """
        Plot error for three configs.
        """

        # Return if figure exists
        path = self.path / 'sl_hyper_comparacao_erro_tres.png'
        if path.exists():
            return

        # Constants
        # cfg_names = ['h_2121', 'h_2421', 'h_2821']
        cfg_names = ['h_2421', 'h_2411', 'h_2322']

        # Constants
        n_rows = 3
        n_cols = 1

        # Create figure
        fig, axs = plt.subplots(n_rows, n_cols, sharey='row', sharex='col', figsize=(12, 9.6))
        axs = axs.flatten()

        # Loop through configs and axes
        for i, (cfg_name, ax) in enumerate(zip(cfg_names, axs)):

            # Get stats
            cfg = self._get_config(cfg_name)
            stats = SL(cfg).stats

            # Plot
            ax.plot(range(1, 50), stats.val_recon_loss, label='Erro de reconstrução')
            ax2 = ax.twinx()
            ax2.plot(range(1, 50), stats.val_kl_loss, color='green', label='Erro de divergência KL')

            # Change axis color
            ax.yaxis.set_tick_params(labelcolor='blue')
            ax2.yaxis.set_tick_params(labelcolor='green')

            # Legend
            lh1, l1 = ax.get_legend_handles_labels()
            lh2, l2 = ax2.get_legend_handles_labels()
            ax.legend([lh1[0]]+[lh2[0]], [l1[0]]+[l2[0]])

            # Axes labels
            if (i % (n_rows * n_cols)) >= (n_cols * n_rows - n_cols):
                ax.set_xlabel('Época')

            # Title
            ax.set_title(cfg.general.config_name_or_prefix + ' (' + r'$\beta=$' + str(cfg.sl.loss.beta) + ')', )

        # Save figure
        fig.savefig(str(path), dpi=600, bbox_inches='tight')

    def _get_config(self, name: str):
        """
        Get config by name.
        :param name: name of config
        :return: master config
        """

        for cfg in self.cfgs:
            if cfg.general.config_name_or_prefix == name:
                return cfg.copy(deep=True)

    def _get_values_and_keys(self, dic: dict, name: str, values: list, keys: list):
        """
        Get attributes names and values.
        :param dic: hyperparameter config dict
        :param name: name of current attribute
        :param values: list of attributes values
        :param keys: list of keys names
        """

        for k, v in dic.items():
            new_name = name + '-' + k if name != "" else k
            if isinstance(v, dict):
                self._get_values_and_keys(v, new_name, values, keys)
            elif v is not None:
                values.append(v)
                keys.append(new_name)

    def _setup(self):
        """
        Set up.
        """

        self._setup_prefix()
        self._setup_alg_dict_paths()
        self._setup_save_config_to_file()
        self._setup_get_params()
        self._setup_remove_invalid_configs()

    def _setup_ignore_static_warnings(self):
        """
        Avoid static method warnings.
        """

        pass

    def _setup_prefix(self):
        """
        Set attribute that contains the name of hyperparameter search config.
        """

        self.prefix = self.cfg.general.config_name_or_prefix

    def _setup_alg_dict_paths(self):
        """
        Set up dict of params to change in the chosen algorithm.
        """

        match self.cfg.general.algorithm:
            case 'rl':
                self.algorithm = RL
                self.hyper_dict = self.cfg.hyperparams.rl.dict()
                self.path = Paths().experiments_rl / self.prefix
            case 'rrt':
                self.algorithm = RRT
                self.hyper_dict = self.cfg.hyperparams.rrt.dict()
                self.path = Paths().experiments_rrt / self.prefix
            case 'sl':
                self.algorithm = SL
                self.hyper_dict = self.cfg.hyperparams.sl.dict()
                self.path = Paths().experiments_sl / self.prefix
            case _:
                raise NotImplementedError

        # Create dir if not exist
        if not self.path.exists():
            self.path.mkdir()

    def _setup_get_params(self):
        """
        Get parameters from config.
        """

        # Get list of keys and values
        values, keys = [], []
        self._get_values_and_keys(self.hyper_dict, "", values, keys)

        # Get permutation of the indexes (suffixes)
        count = []
        for val in values:
            count.append(list(range(1, len(val) + 1)))

        # 'h_111' or 'h_010101' (if there are more than 9 values for any parameter)
        if not any([max(l) > 9 for l in count]):
            suffixes = [''.join([str(n) for n in item]) for item in product(*count)]
        else:
            suffixes = [''.join([str(n).zfill(2) for n in item]) for item in product(*count)]

        # Also permute the values
        values = list(product(*values))

        # Save values, suffixes and keys
        self.values = values
        self.suffixes = suffixes
        self.keys = keys

        # Set attributes in config and set new config name
        cfgs = []
        for value, suffix in zip(values, suffixes):

            # Create copy of original parameters
            cfg_copy = self.cfg.copy(deep=True)

            # Change experiment name
            cfg_copy.general.config_name_or_prefix = self.prefix + '_' + suffix
            cfg_copy.general.description = f"{keys} / {suffix} / {value}"

            # Change hyperparams attributes
            for attr, val in zip(keys, value):

                # Get alias to algorithm level config
                match self.cfg.general.algorithm:
                    case 'rl':
                        cfg_copy_alg = cfg_copy.rl
                    case 'rrt':
                        cfg_copy_alg = cfg_copy.rrt
                    case 'sl':
                        cfg_copy_alg = cfg_copy.sl
                    case _:
                        raise NotImplementedError

                # Get each key in nested key name to access last level of config
                key_chain = attr.split('-')

                # Access attribute
                for key in key_chain[:-1]:
                    cfg_copy_alg = cfg_copy_alg.__getattribute__(key)
                cfg_copy_alg.__setattr__(key_chain[-1], val)

            # Add to list of configs
            cfgs.append(cfg_copy)

        # Add list of configs as attribute
        self.cfgs = cfgs

    def _setup_remove_invalid_configs(self):
        """
        Remove invalid configs, for example, RRT with MPC as controller and rl distance function.
        """

        # Initialize new lists
        cfgs, values, suffixes = [], [], []

        # Loop current configs
        for i, cfg in enumerate(self.cfgs):

            # Invalid configs by algorithm
            match self.cfg.general.algorithm:

                # Invalid configs for RRT
                case 'rrt':

                    # Can't have MPC without RS
                    if cfg.rrt.general.controller == 'mpc' and cfg.rrt.general.distance_fn != 'rs':
                        continue

                    # Don't use RS with RL
                    if cfg.rrt.general.controller == 'rl' and cfg.rrt.general.distance_fn == 'rs':
                        continue

            # Add to new lists
            cfgs.append(self.cfgs[i])
            values.append(self.values[i])
            suffixes.append(self.suffixes[i])

        # Replace
        self.cfgs = cfgs
        self.values = values
        self.suffixes = suffixes

    def _setup_save_config_to_file(self):
        """
        Save config to file.
        """

        # Save config at experiment / algorithm / experiment name folder
        path = self.path / 'config.yaml'
        if not path.exists():
            with open(str(path), 'w') as f:
                yaml.dump(self.cfg.dict(), f)



from collections import OrderedDict

import numpy as np
from gym import spaces

from . import VecEnv


class DummyVecEnv(VecEnv):
  """
    Creates a simple vectorized wrapper for multiple environments

    :param env_fns: ([Gym Environment]) the list of environments to vectorize
    """

  def __init__(self, env_fns):
    self.envs = [fn() for fn in env_fns]
    env = self.envs[0]
    VecEnv.__init__(self, len(env_fns), env.observation_space, env.action_space)
    shapes, dtypes = {}, {}
    self.keys = []
    obs_space = env.observation_space

    self.dummy_env = env
    if isinstance(obs_space, spaces.Dict):
      assert isinstance(obs_space.spaces, OrderedDict)
      subspaces = obs_space.spaces
      if env.compute_reward is not None:
        self.compute_reward = env.compute_reward
      if hasattr(env, 'goal_extraction_function') and env.goal_extraction_function is not None:
        self.goal_extraction_function = env.goal_extraction_function
    else:
      subspaces = {None: obs_space}

    for key, box in subspaces.items():
      shapes[key] = box.shape
      dtypes[key] = box.dtype
      self.keys.append(key)

    self.buf_obs = {k: np.zeros((self.num_envs, ) + tuple(shapes[k]), dtype=dtypes[k]) for k in self.keys}
    self.buf_dones = np.zeros((self.num_envs, ), dtype=np.bool)
    self.buf_rews = np.zeros((self.num_envs, ), dtype=np.float32)
    self.buf_infos = [{} for _ in range(self.num_envs)]
    self.actions = None

  def step_async(self, actions):
    self.actions = actions

  def step_wait(self):
    for env_idx in range(self.num_envs):
      obs, self.buf_rews[env_idx], self.buf_dones[env_idx], self.buf_infos[env_idx] =\
          self.envs[env_idx].step(self.actions[env_idx])
      if self.buf_dones[env_idx]:
        obs = self.envs[env_idx].reset()
      self._save_obs(env_idx, obs)
    if self.keys == [None]:
      return (np.copy(self._obs_from_buf()), np.copy(self.buf_rews), np.copy(self.buf_dones), self.buf_infos.copy())
    else:
      return ({k: np.copy(v)
               for k, v in self._obs_from_buf().items()}, np.copy(self.buf_rews), np.copy(self.buf_dones),
              self.buf_infos.copy())

  def reset(self):
    for env_idx in range(self.num_envs):
      obs = self.envs[env_idx].reset()
      self._save_obs(env_idx, obs)
    if self.keys == [None]:
      return np.copy(self._obs_from_buf())
    else:
      return {k: np.copy(v) for k, v in self._obs_from_buf().items()}

  def close(self):
    return

  def get_images(self):
    return [env.render(mode='rgb_array') for env in self.envs]

  def render(self, *args, **kwargs):
    if self.num_envs == 1:
      return self.envs[0].render(*args, **kwargs)
    else:
      return super().render(*args, **kwargs)

  def _save_obs(self, env_idx, obs):
    for key in self.keys:
      if key is None:
        self.buf_obs[key][env_idx] = obs
      else:
        self.buf_obs[key][env_idx] = obs[key]

  def _obs_from_buf(self):
    if self.keys == [None]:
      return self.buf_obs[None]
    else:
      return self.buf_obs

  def get_attr(self, attr_name, indices=None):
      """Return attribute from vectorized environment (see base class)."""
      target_envs = self._get_target_envs(indices)
      return [getattr(env_i, attr_name) for env_i in target_envs]

  def set_attr(self, attr_name, value, indices=None):
      """Set attribute inside vectorized environments (see base class)."""
      target_envs = self._get_target_envs(indices)
      for env_i in target_envs:
          setattr(env_i, attr_name, value)

  def env_method(self, method_name, *method_args, indices=None, **method_kwargs):
      """Call instance methods of vectorized environments."""
      target_envs = self._get_target_envs(indices)
      return [getattr(env_i, method_name)(*method_args, **method_kwargs) for env_i in target_envs]

  def _get_target_envs(self, indices):
      indices = self._get_indices(indices)
      return [self.envs[i] for i in indices]


import numpy as np
import pandas as pd
import torch as t
from coral_pytorch.dataset import corn_label_from_logits
from torch.nn import functional as f

from rrt_ml.utilities.configs import *
from rrt_ml.utilities.paths import *


class CVAE(t.nn.Module):

    def __init__(self, cfg: MasterConfig):
        """
        Initialize.
        :param cfg: SL config.
        """

        super(CVAE, self).__init__()

        # Attributes
        self.cfg = cfg

        self.device = None
        self.mlp_encoder_shared = None
        self.mlp_encoder_mu = None
        self.mlp_encoder_sigma = None
        self.mlp_decoder = None
        self.beta_norm = None

        self._setup()

    def forward(self, x, y):
        """
        Forward.
        :param x: state vector.
        :param y: conditions vector.
        :return: x_hat, mu and log_sigma.
        """

        # Concatenate input and condition passing through encoder MLP
        x_and_y = t.cat((x, y), dim=1)
        encoder_shared_out = self.mlp_encoder_shared(x_and_y)

        # Get mu and log sigma outputs
        mu = self.mlp_encoder_mu(encoder_shared_out)
        log_sigma = self.mlp_encoder_sigma(encoder_shared_out)

        # Transform log_sigma to sigma
        sigma = t.exp(log_sigma / 2)

        # Re-parametrization trick
        epsilon = t.randn_like(mu)
        z = mu + epsilon * sigma

        # Concatenate latent vector and condition
        z_and_y = t.cat((z, y), dim=1)

        # Pass through decoder MLP
        x_hat = self.mlp_decoder(z_and_y)

        return x_hat, mu, log_sigma

    def loss(self, x_recon, x_real, mu, log_sigma):
        """
        Calculate loss.
        """

        # Individual losses
        kl_loss = self.beta_norm * t.mean(
            t.sum(t.exp(log_sigma) + mu ** 2 - 1 - log_sigma, dim=1), dim=0
        )
        recon_loss = f.mse_loss(x_recon, x_real)

        # Weighted loss
        loss = kl_loss + recon_loss

        return loss, kl_loss, recon_loss

    def _setup(self):
        """
        Setup model.
        """

        self._setup_device()
        self._setup_encoder()
        self._setup_decoder()
        self._setup_beta()
        self._setup_float()

    def _setup_device(self):
        """
        Setup device, GPU if cuda available.
        """

        # Check if cuda available
        self.device = "cuda" if t.cuda.is_available() else "cpu"

    def _setup_encoder(self):
        """
        Setup encoder MLP.
        """

        # Encoder MLP
        mlp_encoder = []
        for i, n_units in enumerate(self.cfg.sl.net.mlp_units):
            mlp_encoder.extend([t.nn.LazyLinear(n_units), t.nn.ReLU()])
        self.mlp_encoder_shared = t.nn.Sequential(*mlp_encoder)

        # Encoder mu and sigma
        self.mlp_encoder_mu = t.nn.LazyLinear(self.cfg.sl.dim.latent)
        self.mlp_encoder_sigma = t.nn.LazyLinear(self.cfg.sl.dim.latent)

    def _setup_decoder(self):
        """
        Setup decoder MLP.
        """

        # Decoder
        mlp_decoder = []
        for i, n_units in enumerate(self.cfg.sl.net.mlp_units):
            mlp_decoder.extend([t.nn.LazyLinear(n_units), t.nn.ReLU()])
        mlp_decoder.append(t.nn.LazyLinear(self.cfg.sl.dim.state))
        self.mlp_decoder = t.nn.Sequential(*mlp_decoder)

    def _setup_beta(self):
        """
        Setup KL balancing.
        """

        self.beta_norm = self.cfg.sl.loss.beta * self.cfg.sl.dim.latent / self.cfg.sl.dim.state

    def _setup_float(self):
        """
        Change model parameters to 32 bit.
        """

        self.float()


class Ordinal(t.nn.Module):

    def __init__(self, cfg: MasterConfig = MasterConfig()):
        """
        Initialize ordinal MLP regressor.
        """

        super().__init__()

        self.cfg = cfg
        self.timesteps_range = None  # type: None | list[int, int]

        self._setup()

    def forward(self, x):
        """
        Forward.
        :param x: input.
        :return: output.
        """

        return self.net(x)

    def get_distance(self, x: np.ndarray):
        """
        Predict distance.
        :return: distance predictions.
        """

        # Convert to tensor
        x_t = t.tensor(x, dtype=t.float32).to("cuda")

        # Forward
        with t.no_grad():
            self.eval()
            y = self(x_t)
            y = corn_label_from_logits(y)

        # Add minimum timesteps from dataset to get actual values
        pred = y.cpu().numpy().flatten() - self.timesteps_range[0]

        return pred

    def _setup(self):
        """
        Set up.
        """

        self._setup_timesteps_range()
        self._setup_net()
        self._setup_cuda()

    def _setup_cuda(self):
        """
        Set model to cuda.
        """

        self.cuda()

    def _setup_timesteps_range(self):
        """
        Get min max timesteps from data.
        """

        # Set column names
        col_names = ['x', 'y', 'sin_theta', 'cos_theta', 'success', 'time', 'timesteps', 'length']

        # Read data
        path = Paths().data_rl_distance / self.cfg.rrt.names.rl / 'train.csv'
        df = pd.read_csv(path, names=col_names, index_col=None)

        # Get min and max values for timesteps
        self.timesteps_range = [df['timesteps'].min(), df['timesteps'].max()]

    def _setup_net(self):
        """
        Set up net architecture.
        """

        # Get number of classes
        max_timesteps = int(self.timesteps_range[1] - self.timesteps_range[0] + 1)

        # Net
        units = []
        for i, n_units in enumerate(self.cfg.rrt.ordinal.layers):
            units.extend([t.nn.LazyLinear(n_units), t.nn.GELU()])
        units.append(t.nn.LazyLinear(max_timesteps))
        self.net = t.nn.Sequential(*units)

        # Convert to float
        self.float()
        self.net.float()


class OrdinalEnsemble:

    def __init__(self, cfg: MasterConfig = MasterConfig()):
        """
        Initialize.
        :param cfg: master config.
        """

        self.models = []

    def get_distance(self, x):
        """
        Get distance prediction from ensemble.
        :param x: input array [[x_car y_car sin_car cos_car]]
        :return: distance prediction
        """

        # Convert to tensor
        x_t = t.tensor(x, dtype=t.float32, device='cuda', requires_grad=False)

        # Initialize array
        batch_size = x_t.size()[0]
        all_pred = np.zeros(shape=(20, batch_size))

        # Forward
        with t.no_grad():
            for i, model in enumerate(self.models):
                model.eval()
                all_pred[i, :] = corn_label_from_logits(model(x_t)).detach().cpu().numpy()

        # Get mean value
        all_pred = np.mean(all_pred, axis=0)

        # Add minimum timesteps from dataset to get actual values
        all_pred = all_pred - self.models[0].timesteps_range[0]

        return all_pred



"""
A generic pytorch VAE. Based on https://github.com/pytorch/examples/blob/master/vae/main.py.

NOTE: NOT TESTED/USED may be wrong.
"""

import torch
from torch import nn, optim
from torch.nn import functional as F

class VAE(nn.Module):
  def __init__(self, input_dim, hidden_dim=256, latent_dim=64, num_hidden=2):
    super(VAE, self).__init__()

    self.input_dim = input_dim
    self.hidden_dim = hidden_dim
    self.latent_dim = latent_dim
    self.num_hidden = 2

    layer_sizes = [input_dim] + [hidden_dim] * num_hidden
    self.encoder = []
    for i, o in zip(layer_sizes[:-1], layer_sizes[1:]):
      self.encoder.append(nn.Linear(i, o))
      self.encoder.append(nn.ReLU())
    self.encoder = nn.Sequential(*self.encoder)

    self.mu = nn.Linear(hidden_dim, latent_dim)
    self.sigma = nn.Linear(hidden_dim, latent_dim)

    layer_sizes = [latent_dim] + [hidden_dim] * num_hidden + [input_dim]
    self.decoder = []
    for i, o in zip(layer_sizes[:-1], layer_sizes[1:]):
      self.decoder.append(nn.Linear(i, o))
      self.decoder.append(nn.ReLU())
    self.decoder.pop()
    self.decoder = nn.Sequential(*self.decoder)

  def encode(self, x):
    h = self.encoder(x)
    return self.mu(h), self.sigma(h)

  def reparameterize(self, mu, logvar):
    std = torch.exp(0.5*logvar)
    eps = torch.randn_like(std)
    return mu + eps*std

  def decode(self, z):
    return self.decoder(z)

  def forward(self, x):
    mu, logvar = self.encode(x)
    z = self.reparameterize(mu, logvar)
    return self.decode(z), mu, logvar


# Mean reconstruction + KL divergence losses summed over all elements and batch
def vae_loss_function(recon_x, x, mu, logvar):
    MSE = torch.sum((recon_x - x)**2, -1)

    # see Appendix B from VAE paper:
    # Kingma and Welling. Auto-Encoding Variational Bayes. ICLR, 2014
    # https://arxiv.org/abs/1312.6114
    # 0.5 * sum(1 + log(sigma^2) - mu^2 - sigma^2)
    KLD = -0.5 * torch.sum(1 + logvar - mu.pow(2) - logvar.exp(), -1)

    return (MSE + KLD).mean()


import mrl
from typing import Optional, Callable, List, Union
from mrl.utils.misc import AttrDict, set_global_seeds
import numpy as np

import gym

try:
  from baselines.common.atari_wrappers import make_atari, wrap_deepmind
except:
  pass
from mrl.utils.vec_env.subproc_vec_env import SubprocVecEnv
from mrl.utils.vec_env.dummy_vec_env import DummyVecEnv

import time


class EnvModule(mrl.Module):
  """
  Used to wrap state-less environments in an mrl.Module.
  Vectorizes the environment.
  """
  def __init__(
      self,
      env: Union[str, Callable],
      num_envs: int = 1,
      seed: Optional[int] = None,
      name: Optional[str] = None,
      episode_life=True  # for Atari
  ):

    super().__init__(name or 'env', required_agent_modules=[], locals=locals())

    self.num_envs = num_envs

    if seed is None:
      seed = int(time.time())

    if isinstance(env, str):
      sample_env = make_env_by_id(env, seed, 0, episode_life)()
      env_list = [make_env_by_id(env, seed, i, episode_life) for i in range(num_envs)]
    else:
      sample_env = make_env(env, seed, 0)()
      assert isinstance(sample_env, gym.core.Env), "Only gym environments supported for now!"
      env_list = [make_env(env, seed, i) for i in range(num_envs)]

    if num_envs == 1:
      self.env = DummyVecEnv(env_list)
    else:
      self.env = SubprocVecEnv(env_list)
    print('Initializing env!')

    self.render = self.env.render
    self.observation_space = sample_env.observation_space
    self.action_space = sample_env.action_space

    if isinstance(self.action_space, gym.spaces.Discrete):
      self.action_dim = self.action_space.n
      self.max_action = None
    else:
      assert isinstance(self.action_space, gym.spaces.Box), "Only Box/Discrete actions supported for now!"
      self.action_dim = self.action_space.shape[0]
      self.max_action = self.action_space.high[0]
      assert np.allclose(self.action_space.high,
                         -self.action_space.low), "Action high/lows must equal! Several modules rely on this"

    self.goal_env = False
    self.goal_dim = 0

    if isinstance(self.observation_space, gym.spaces.Dict):
      self.goal_env = True
      self.compute_reward = sample_env.compute_reward
      self.state_dim = int(np.prod(self.env.observation_space['observation'].shape))
      self.goal_dim = int(np.prod(self.env.observation_space['desired_goal'].shape))
    else:
      self.state_dim = int(np.prod(self.env.observation_space.shape))

    self.state = self.env.reset()

  def step(self, action):
    res = self.env.step(action)
    self.state = res[0]
    return res

  def reset(self, indices=None):
    if not indices:
      self.state = self.env.reset()
      return self.state
    else:
      reset_states = self.env.env_method('reset', indices=indices)
      if self.goal_env:
        for i, reset_state in zip(indices, reset_states):
          for key in reset_state:
            self.state[key][i] = reset_state[key]
      else:
        for i, reset_state in zip(indices, reset_states):
          self.state[i] = reset_state
      return self.state


def make_env(env_fn, seed, rank):
  """
  Utility function for multiprocessed env.
  
  :param env_id: (str) the environment ID
  :param num_env: (int) the number of environment you wish to have in subprocesses
  :param seed: (int) the inital seed for RNG
  """
  def _init():
    env = env_fn()
    env.seed(seed + rank)
    env = ReturnAndObsWrapper(env)
    return env

  set_global_seeds(seed)
  return _init


### BELOW is based on https://github.com/ShangtongZhang/DeepRL/blob/master/deep_rl/component/envs.py
### Added a fix for the VecEnv bug in infinite-horizon envs.


# adapted from https://github.com/ikostrikov/pytorch-a2c-ppo-acktr/blob/master/envs.py
def make_env_by_id(env_id, seed, rank, episode_life=True):
  """Used for regular gym environments and Atari Envs"""
  def _init():
    if env_id.startswith("dm"):
      import dm_control2gym
      _, domain, task = env_id.split('-')
      env = dm_control2gym.make(domain_name=domain, task_name=task)
    else:
      env = gym.make(env_id)
    is_atari = hasattr(gym.envs, 'atari') and isinstance(env.unwrapped, gym.envs.atari.atari_env.AtariEnv)
    if is_atari:
      env = make_atari(env_id)
    env.seed(seed + rank)
    if is_atari:
      env = wrap_deepmind(env, episode_life=episode_life, clip_rewards=False, frame_stack=False, scale=False)
      obs_shape = env.observation_space.shape
      if len(obs_shape) == 3:
        env = TransposeImage(env)
      env = FrameStack(env, 4)
    env = ReturnAndObsWrapper(env)
    return env

  set_global_seeds(seed)
  return _init


class ReturnAndObsWrapper(gym.Wrapper):
  def __init__(self, env):
    gym.Wrapper.__init__(self, env)
    self.total_rewards = 0

  def step(self, action):
    obs, reward, done, info = self.env.step(action)
    info = AttrDict(info)
    self.total_rewards += reward
    if done:
      info.done_observation = obs
      info.terminal_state = True
      if info.get('TimeLimit.truncated'):
        info.terminal_state = False
      info.episodic_return = self.total_rewards
      self.total_rewards = 0
    else:
      info.terminal_state = False
      info.episodic_return = None
    return obs, reward, done, info

  def render(self, *args, **kwargs):
    return self.env.render(*args, **kwargs)

  def reset(self):
    return self.env.reset()

  def __getattr__(self, attr):
    return getattr(self.env, attr)


class FirstVisitDoneWrapper(gym.Wrapper):
  """A wrapper for sparse reward goal envs that makes them terminate
  upon achievement"""
  def step(self, action):
    obs, reward, done, info = self.env.step(action)
    if np.allclose(reward, 0.):
      done = True
      info['is_success'] = True
      if info.get('TimeLimit.truncated'):
        del info['TimeLimit.truncated']
    return obs, reward, done, info

  def reset(self):
    return self.env.reset()

  def __getattr__(self, attr):
    return getattr(self.env, attr)


class TransposeImage(gym.ObservationWrapper):
  def __init__(self, env=None):
    super(TransposeImage, self).__init__(env)
    obs_shape = self.observation_space.shape
    self.observation_space = gym.spaces.Box(self.observation_space.low[0, 0, 0],
                                            self.observation_space.high[0, 0, 0],
                                            [obs_shape[2], obs_shape[1], obs_shape[0]],
                                            dtype=self.observation_space.dtype)

  def observation(self, observation):
    return observation.transpose(2, 0, 1)


# The original LayzeFrames doesn't work well
class LazyFrames(object):
  def __init__(self, frames):
    """This object ensures that common frames between the observations are only stored once.
      It exists purely to optimize memory usage which can be huge for DQN's 1M frames replay
      buffers.
      This object should only be converted to numpy array before being passed to the model.
      You'd not believe how complex the previous solution was."""
    self._frames = frames

  def __array__(self, dtype=None):
    out = np.concatenate(self._frames, axis=0)
    if dtype is not None:
      out = out.astype(dtype)
    return out

  def __len__(self):
    return len(self.__array__())

  def __getitem__(self, i):
    return self.__array__()[i]


class FrameStack(gym.Wrapper):
  def __init__(self, env, k):
    """Stack k last frames.

    Returns lazy array, which is much more memory efficient.

    See Also
    --------
    baselines.common.atari_wrappers.LazyFrames
    """
    gym.Wrapper.__init__(self, env)
    self.k = k
    self.frames = deque([], maxlen=k)
    shp = env.observation_space.shape
    self.observation_space = spaces.Box(low=0, high=255, shape=(shp[0], shp[1], shp[2] * k), dtype=np.uint8)

  def reset(self):
    ob = self.env.reset()
    for _ in range(self.k):
      self.frames.append(ob)
    return self._get_ob()

  def step(self, action):
    ob, reward, done, info = self.env.step(action)
    self.frames.append(ob)
    return self._get_ob(), reward, done, info

  def _get_ob(self):
    assert len(self.frames) == self.k
    return LazyFrames(list(self.frames))



"""
This is a fixed horizon DDPG, based on https://arxiv.org/pdf/1907.04543.pdf.
"""
from mrl.algorithms.continuous_off_policy import *

class FHDDPG(OffPolicyActorCritic):

  def optimize_from_batch(self, states, actions, rewards, next_states, gammas): 
    config = self.config

    a_next_max = self.actor_target(next_states)
    noise = torch.randn_like(a_next_max) * (config.td3_noise * self.action_scale)
    noise = noise.clamp(-config.td3_noise_clip * self.action_scale,
                        config.td3_noise_clip * self.action_scale)
    a_next_max = (a_next_max + noise).clamp(-self.action_scale,
                                            self.action_scale)

    q_next = self.critic_target(next_states, a_next_max) # batch x n_horizons
    q_next = F.pad(q_next, (1, 0))[:, :-1]

    target = (rewards + gammas * q_next)
    target = torch.clamp(target, *self.config.clip_target_range).detach()

    if hasattr(self, 'logger') and self.config.opt_steps % 100 == 0:
      self.logger.add_histogram('Optimize/First_horizon_target', target[:,0])
      self.logger.add_histogram('Optimize/Last_horizon_target', target[:,-1])
    
    q = self.critic(states, actions)
    critic_loss = F.mse_loss(q, target)

    self.critic_opt.zero_grad()
    critic_loss.backward()
    self.critic_opt.step()

    a = self.actor(states) + noise
    actor_loss = -self.critic(states, a)[:, -5:].mean() # update actor using ensemble of last 5 horizons
    if self.config.action_l2_regularization:
      actor_loss += self.config.action_l2_regularization * F.mse_loss(a, torch.zeros_like(a))

    self.actor_opt.zero_grad()
    actor_loss.backward()
    self.actor_opt.step()


"""
  Adversarial training is able to improve the performance of an ensemble by
  treating adversarial samples as the augmented training data. The fast
  gradient sign method (FGSM) is used to generate adversarial samples.

  Reference:
      B. Lakshminarayanan, A. Pritzel, C. Blundell., Simple and Scalable
      Predictive Uncertainty Estimation using Deep Ensembles, NIPS 2017.
"""


import torch
import torch.nn as nn
import torch.nn.functional as F

import warnings
from joblib import Parallel, delayed

from ._base import BaseModule, BaseClassifier, BaseRegressor
from ._base import torchensemble_model_doc
from .utils import io
from .utils import set_module
from .utils import operator as op


__all__ = ["AdversarialTrainingClassifier", "AdversarialTrainingRegressor"]


__fit_doc = """
    Parameters
    ----------
    train_loader : torch.utils.data.DataLoader
        A :mod:`torch.utils.data.DataLoader` container that contains the
        training data.
    epochs : int, default=100
        The number of training epochs.
    epsilon : float, default=0.01
        The step used to generate adversarial samples in the fast gradient
        sign method (FGSM), which should be in the range [0, 1].
    log_interval : int, default=100
        The number of batches to wait before logging the training status.
    test_loader : torch.utils.data.DataLoader, default=None
        A :mod:`torch.utils.data.DataLoader` container that contains the
        evaluating data.

        - If ``None``, no validation is conducted after each training
          epoch.
        - If not ``None``, the ensemble will be evaluated on this
          dataloader after each training epoch.
    save_model : bool, default=True
        Specify whether to save the model parameters.

        - If test_loader is ``None``, the ensemble fully trained will be
          saved.
        - If test_loader is not ``None``, the ensemble with the best
          validation performance will be saved.
    save_dir : string, default=None
        Specify where to save the model parameters.

        - If ``None``, the model will be saved in the current directory.
        - If not ``None``, the model will be saved in the specified
          directory: ``save_dir``.
"""


def _adversarial_training_model_doc(header, item="fit"):
    """
    Decorator on obtaining documentation for different adversarial training
    models.
    """

    def get_doc(item):
        """Return selected item"""
        __doc = {"fit": __fit_doc}
        return __doc[item]

    def adddoc(cls):
        doc = [header + "\n\n"]
        doc.extend(get_doc(item))
        cls.__doc__ = "".join(doc)
        return cls

    return adddoc


def _parallel_fit_per_epoch(
    train_loader,
    epsilon,
    estimator,
    cur_lr,
    optimizer,
    criterion,
    idx,
    epoch,
    log_interval,
    device,
    is_classification,
):
    """
    Private function used to fit base estimators in parallel.

    WARNING: Parallelization when fitting large base estimators may cause
    out-of-memory error.
    """

    if cur_lr:
        # Parallelization corrupts the binding between optimizer and scheduler
        set_module.update_lr(optimizer, cur_lr)

    for batch_idx, elem in enumerate(train_loader):

        data, target = io.split_data_target(elem, device)
        batch_size = data[0].size(0)
        for tensor in data:
            tensor.requires_grad = True

        # Get adversarial samples
        _output = estimator(*data)
        _loss = criterion(_output, target)
        _loss.backward()
        data_grad = [tensor.grad.data for tensor in data]
        adv_data = _get_fgsm_samples(data, epsilon, data_grad)

        # Compute the training loss
        optimizer.zero_grad()
        org_output = estimator(*data)
        adv_output = estimator(*adv_data)
        loss = criterion(org_output, target) + criterion(adv_output, target)
        loss.backward()
        optimizer.step()

        # Print training status
        if batch_idx % log_interval == 0:

            # Classification
            if is_classification:
                _, predicted = torch.max(org_output.data, 1)
                correct = (predicted == target).sum().item()

                msg = (
                    "Estimator: {:03d} | Epoch: {:03d} | Batch: {:03d}"
                    " | Loss: {:.5f} | Correct: {:d}/{:d}"
                )
                print(
                    msg.format(
                        idx, epoch, batch_idx, loss, correct, batch_size
                    )
                )
            # Regression
            else:
                msg = (
                    "Estimator: {:03d} | Epoch: {:03d} | Batch: {:03d}"
                    " | Loss: {:.5f}"
                )
                print(msg.format(idx, epoch, batch_idx, loss))

    return estimator, optimizer


def _get_fgsm_samples(sample_list, epsilon, sample_grad_list):
    """
    Private functions used to generate adversarial samples with fast gradient
    sign method (FGSM).
    """

    perturbed_sample_list = []
    for sample, sample_grad in zip(sample_list, sample_grad_list):
        # Check the input range of `sample`
        min_value, max_value = torch.min(sample), torch.max(sample)
        if not 0 <= min_value < max_value <= 1:
            msg = (
                "The input range of samples passed to adversarial training"
                " should be in the range [0, 1], but got [{:.3f}, {:.3f}]"
                " instead."
            )
            raise ValueError(msg.format(min_value, max_value))

        sign_sample_grad = sample_grad.sign()
        perturbed_sample = sample + epsilon * sign_sample_grad
        perturbed_sample = torch.clamp(perturbed_sample, 0, 1)

        perturbed_sample_list.append(perturbed_sample)

    return perturbed_sample_list


class _BaseAdversarialTraining(BaseModule):
    def _validate_parameters(self, epochs, epsilon, log_interval):
        """Validate hyper-parameters on training the ensemble."""

        if not epochs > 0:
            msg = (
                "The number of training epochs = {} should be strictly"
                " positive."
            )
            self.logger.error(msg.format(epochs))
            raise ValueError(msg.format(epochs))

        if not 0 < epsilon <= 1:
            msg = (
                "The step used to generate adversarial samples in FGSM"
                " should be in the range (0, 1], but got {} instead."
            )
            self.logger.error(msg.format(epsilon))
            raise ValueError(msg.format(epsilon))

        if not log_interval > 0:
            msg = (
                "The number of batches to wait before printting the"
                " training status should be strictly positive, but got {}"
                " instead."
            )
            self.logger.error(msg.format(log_interval))
            raise ValueError(msg.format(log_interval))


@torchensemble_model_doc(
    """Implementation on the AdversarialTrainingClassifier.""",  # noqa: E501
    "model",
)
class AdversarialTrainingClassifier(_BaseAdversarialTraining, BaseClassifier):
    @torchensemble_model_doc(
        """Implementation on the data forwarding in AdversarialTrainingClassifier.""",  # noqa: E501
        "classifier_forward",
    )
    def forward(self, *x):
        # Take the average over class distributions from all base estimators.
        outputs = [
            F.softmax(estimator(*x), dim=1) for estimator in self.estimators_
        ]
        proba = op.average(outputs)

        return proba

    @torchensemble_model_doc(
        """Set the attributes on optimizer for AdversarialTrainingClassifier.""",  # noqa: E501
        "set_optimizer",
    )
    def set_optimizer(self, optimizer_name, **kwargs):
        super().set_optimizer(optimizer_name, **kwargs)

    @torchensemble_model_doc(
        """Set the attributes on scheduler for AdversarialTrainingClassifier.""",  # noqa: E501
        "set_scheduler",
    )
    def set_scheduler(self, scheduler_name, **kwargs):
        super().set_scheduler(scheduler_name, **kwargs)

    @torchensemble_model_doc(
        """Set the training criterion for AdversarialTrainingClassifier.""",
        "set_criterion",
    )
    def set_criterion(self, criterion):
        super().set_criterion(criterion)

    @_adversarial_training_model_doc(
        """Implementation on the training stage of AdversarialTrainingClassifier.""",  # noqa: E501
        "fit",
    )
    def fit(
        self,
        train_loader,
        epochs=100,
        epsilon=0.5,
        log_interval=100,
        test_loader=None,
        save_model=True,
        save_dir=None,
    ):

        self._validate_parameters(epochs, epsilon, log_interval)
        self.n_outputs = self._decide_n_outputs(train_loader)

        # Instantiate a pool of base estimators, optimizers, and schedulers.
        estimators = []
        for _ in range(self.n_estimators):
            estimators.append(self._make_estimator())

        optimizers = []
        for i in range(self.n_estimators):
            optimizers.append(
                set_module.set_optimizer(
                    estimators[i], self.optimizer_name, **self.optimizer_args
                )
            )

        if self.use_scheduler_:
            scheduler_ = set_module.set_scheduler(
                optimizers[0], self.scheduler_name, **self.scheduler_args
            )

        # Check the training criterion
        if not hasattr(self, "_criterion"):
            self._criterion = nn.CrossEntropyLoss()

        # Utils
        best_acc = 0.0

        # Internal helper function on pesudo forward
        def _forward(estimators, *x):
            outputs = [
                F.softmax(estimator(*x), dim=1) for estimator in estimators
            ]
            proba = op.average(outputs)

            return proba

        # Maintain a pool of workers
        with Parallel(n_jobs=self.n_jobs) as parallel:

            # Training loop
            for epoch in range(epochs):
                self.train()

                if self.use_scheduler_:
                    cur_lr = scheduler_.get_last_lr()[0]
                else:
                    cur_lr = None

                if self.n_jobs and self.n_jobs > 1:
                    msg = "Parallelization on the training epoch: {:03d}"
                    self.logger.info(msg.format(epoch))

                rets = parallel(
                    delayed(_parallel_fit_per_epoch)(
                        train_loader,
                        epsilon,
                        estimator,
                        cur_lr,
                        optimizer,
                        self._criterion,
                        idx,
                        epoch,
                        log_interval,
                        self.device,
                        False,
                    )
                    for idx, (estimator, optimizer) in enumerate(
                        zip(estimators, optimizers)
                    )
                )

                estimators, optimizers = [], []
                for estimator, optimizer in rets:
                    estimators.append(estimator)
                    optimizers.append(optimizer)

                # Validation
                if test_loader:
                    self.eval()
                    with torch.no_grad():
                        correct = 0
                        total = 0
                        for _, elem in enumerate(test_loader):
                            data, target = io.split_data_target(
                                elem, self.device
                            )
                            output = _forward(estimators, *data)
                            _, predicted = torch.max(output.data, 1)
                            correct += (predicted == target).sum().item()
                            total += target.size(0)
                        acc = 100 * correct / total

                        if acc > best_acc:
                            best_acc = acc
                            self.estimators_ = nn.ModuleList()  # reset
                            self.estimators_.extend(estimators)
                            if save_model:
                                io.save(self, save_dir, self.logger)

                        msg = (
                            "Epoch: {:03d} | Validation Acc: {:.3f}"
                            " % | Historical Best: {:.3f} %"
                        )
                        self.logger.info(msg.format(epoch, acc, best_acc))
                        if self.tb_logger:
                            self.tb_logger.add_scalar(
                                "adversarial_training/Validation_Acc",
                                acc,
                                epoch,
                            )

                # Update the scheduler
                with warnings.catch_warnings():

                    # UserWarning raised by PyTorch is ignored because
                    # scheduler does not have a real effect on the optimizer.
                    warnings.simplefilter("ignore", UserWarning)

                    if self.use_scheduler_:
                        scheduler_.step()

        self.estimators_ = nn.ModuleList()
        self.estimators_.extend(estimators)
        if save_model and not test_loader:
            io.save(self, save_dir, self.logger)

    @torchensemble_model_doc(item="classifier_evaluate")
    def evaluate(self, test_loader, return_loss=False):
        return super().evaluate(test_loader, return_loss)

    @torchensemble_model_doc(item="predict")
    def predict(self, *x):
        return super().predict(*x)


@torchensemble_model_doc(
    """Implementation on the AdversarialTrainingRegressor.""",  # noqa: E501
    "model",
)
class AdversarialTrainingRegressor(_BaseAdversarialTraining, BaseRegressor):
    @torchensemble_model_doc(
        """Implementation on the data forwarding in AdversarialTrainingRegressor.""",  # noqa: E501
        "regressor_forward",
    )
    def forward(self, *x):
        # Take the average over predictions from all base estimators.
        outputs = [estimator(*x) for estimator in self.estimators_]
        pred = op.average(outputs)

        return pred

    @torchensemble_model_doc(
        """Set the attributes on optimizer for AdversarialTrainingRegressor.""",  # noqa: E501
        "set_optimizer",
    )
    def set_optimizer(self, optimizer_name, **kwargs):
        super().set_optimizer(optimizer_name, **kwargs)

    @torchensemble_model_doc(
        """Set the attributes on scheduler for AdversarialTrainingRegressor.""",  # noqa: E501
        "set_scheduler",
    )
    def set_scheduler(self, scheduler_name, **kwargs):
        super().set_scheduler(scheduler_name, **kwargs)

    @torchensemble_model_doc(
        """Set the training criterion for AdversarialTrainingRegressor.""",
        "set_criterion",
    )
    def set_criterion(self, criterion):
        super().set_criterion(criterion)

    @_adversarial_training_model_doc(
        """Implementation on the training stage of AdversarialTrainingRegressor.""",  # noqa: E501
        "fit",
    )
    def fit(
        self,
        train_loader,
        epochs=100,
        epsilon=0.5,
        log_interval=100,
        test_loader=None,
        save_model=True,
        save_dir=None,
    ):

        self._validate_parameters(epochs, epsilon, log_interval)
        self.n_outputs = self._decide_n_outputs(train_loader)

        # Instantiate a pool of base estimators, optimizers, and schedulers.
        estimators = []
        for _ in range(self.n_estimators):
            estimators.append(self._make_estimator())

        optimizers = []
        for i in range(self.n_estimators):
            optimizers.append(
                set_module.set_optimizer(
                    estimators[i], self.optimizer_name, **self.optimizer_args
                )
            )

        if self.use_scheduler_:
            scheduler_ = set_module.set_scheduler(
                optimizers[0], self.scheduler_name, **self.scheduler_args
            )

        # Check the training criterion
        if not hasattr(self, "_criterion"):
            self._criterion = nn.MSELoss()

        # Utils
        best_loss = float("inf")

        # Internal helper function on pesudo forward
        def _forward(estimators, *x):
            outputs = [estimator(*x) for estimator in estimators]
            pred = op.average(outputs)

            return pred

        # Maintain a pool of workers
        with Parallel(n_jobs=self.n_jobs) as parallel:

            # Training loop
            for epoch in range(epochs):
                self.train()

                if self.use_scheduler_:
                    cur_lr = scheduler_.get_last_lr()[0]
                else:
                    cur_lr = None

                if self.n_jobs and self.n_jobs > 1:
                    msg = "Parallelization on the training epoch: {:03d}"
                    self.logger.info(msg.format(epoch))

                rets = parallel(
                    delayed(_parallel_fit_per_epoch)(
                        train_loader,
                        epsilon,
                        estimator,
                        cur_lr,
                        optimizer,
                        self._criterion,
                        idx,
                        epoch,
                        log_interval,
                        self.device,
                        True,
                    )
                    for idx, (estimator, optimizer) in enumerate(
                        zip(estimators, optimizers)
                    )
                )

                estimators, optimizers = [], []
                for estimator, optimizer in rets:
                    estimators.append(estimator)
                    optimizers.append(optimizer)

                # Validation
                if test_loader:
                    self.eval()
                    with torch.no_grad():
                        val_loss = 0.0
                        for _, elem in enumerate(test_loader):
                            data, target = io.split_data_target(
                                elem, self.device
                            )
                            output = _forward(estimators, *data)
                            val_loss += self._criterion(output, target)
                        val_loss /= len(test_loader)

                        if val_loss < best_loss:
                            best_loss = val_loss
                            self.estimators_ = nn.ModuleList()
                            self.estimators_.extend(estimators)
                            if save_model:
                                io.save(self, save_dir, self.logger)

                        msg = (
                            "Epoch: {:03d} | Validation Loss:"
                            " {:.5f} | Historical Best: {:.5f}"
                        )
                        self.logger.info(
                            msg.format(epoch, val_loss, best_loss)
                        )
                        if self.tb_logger:
                            self.tb_logger.add_scalar(
                                "adversirial_training/Validation_Loss",
                                val_loss,
                                epoch,
                            )

                # Update the scheduler
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", UserWarning)

                    if self.use_scheduler_:
                        scheduler_.step()

        self.estimators_ = nn.ModuleList()
        self.estimators_.extend(estimators)
        if save_model and not test_loader:
            io.save(self, save_dir, self.logger)

    @torchensemble_model_doc(item="regressor_evaluate")
    def evaluate(self, test_loader):
        return super().evaluate(test_loader)

    @torchensemble_model_doc(item="predict")
    def predict(self, *x):
        return super().predict(*x)



import mrl
from mrl.utils.random_process import *

class ContinuousActionNoise(mrl.Module):
  def __init__(self, random_process_cls = GaussianProcess, *args, **kwargs):
    super().__init__('action_noise', required_agent_modules = ['env'], locals=locals())
    self._random_process_cls = random_process_cls
    self._args = args
    self._kwargs = kwargs

  def _setup(self):
    self.random_process = self._random_process_cls((self.env.num_envs, self.env.action_dim,), *self._args, **self._kwargs)
    self.varied_action_noise = self.config.get('varied_action_noise')

  def __call__(self, action):
    factor = 1
    if self.varied_action_noise:
      n_envs = self.env.num_envs
      factor = np.arange(0., 1. + (1./n_envs), 1./(n_envs-1)).reshape(n_envs, 1)
    
    return action + (self.random_process.sample() * self.env.max_action * factor)[:len(action)]

  def save(self, save_folder):
    self._save_props(['random_process'], save_folder)

  def load(self, save_folder):
    self._load_props(['random_process'], save_folder)